from procgraph import Generator, Block, ModelExecutionError, BadConfig
import os
import traceback

__all__ = [
    "TextLog",
]


class TextLog(Generator):
    """ 
        This represents a generic log reader that reads
        from a file line-by-line. 
        
        Subclasses should overload the parse_format() static method.
        Also they should put the adequate output() methods.
    """

    Block.config("file", "File to read. If it ends with ``bz2``, it is treated" " as compressed.")

    def init(self):
        filename = self.config.file
        filename = os.path.expandvars(filename)
        filename = os.path.expanduser(filename)

        if not os.path.exists(filename):
            raise BadConfig("File %r does not exist." % filename, self, "file")

        # TODO: add .gz
        if filename.endswith("bz2"):
            import bz2

            self.stream = bz2.BZ2File(filename)
        else:
            self.stream = open(filename, "r")

        self.state.line = 0  # line counter
        self.read_next_line()

        if self.timestamp is None:
            raise Exception("Empty file %s" % filename)

    def read_next_line(self):
        line = self.state.line
        while True:
            next_line = self.stream.readline()
            # skip empty lines
            if next_line == "\n":
                continue
            else:
                break

        # check end of file
        if not next_line:
            self.timestamp = None
            self.values = None
            return

        # strip newline
        next_line = next_line.rstrip("\n")

        try:
            result_tuple = self.parse_format(next_line)
            # if the line is valid, result_tuple is a tuple
            if result_tuple is not None:
                # TODO: add explicit check
                if not isinstance(result_tuple, tuple) or len(result_tuple) != 2:
                    msg = "Expected tuple of len 2, got %s" % result_tuple
                    raise ValueError(msg)
                self.timestamp, self.values = result_tuple

                if not isinstance(self.values, list):
                    msg = "Expected a list"
                    raise ValueError(self.values)
                for x in self.values:
                    if not isinstance(x, tuple) or len(x) != 2:
                        msg = "Expected a tuple name/value"
                        raise ValueError(self.values)

                self.state.line = line + 1
            else:
                # If the result is None, it means the line did not contain
                # data relevant for us; skip to the next one
                self.state.line = line + 1
                self.read_next_line()
        except Exception as e:
            traceback.print_exc()
            msg = "While reading line %s of file %s (='%s'): %s" % (
                line,
                self.get_config("file"),
                next_line,
                e,
            )
            raise ModelExecutionError(msg, self)

    def next_data_status(self):
        # TODO: put new interface
        if self.timestamp is None:  # EOF
            return (False, None)
        else:
            return (True, self.timestamp)

    def update(self):
        for signal, value in self.values:
            self.set_output(signal, value, timestamp=self.timestamp)

        self.read_next_line()

    def parse_format(self, line):
        """ 
            Function implemented by subclasses to interpret one line from 
            the log.
        
            Return None is the line should be ignored (for example it is 
            a comment). Otherwise, return a tuple (timestamp, array of 
            (name, value) ).
        """
        raise ValueError("Function parse_format() not implemented.")



import numpy
from procgraph import simple_block, BadConfig


@simple_block
def choose(rgb, channel=0):
    """ Chooses a channel from an rgb image and replicates it
        over the other two.

        Raises an error if ``channel`` is invalid.

        :param rgb: Input image.
        :param channel: Which channel to choose.
        :return: processed: The processed image.
    """
    if not channel in [0, 1, 2]:
        raise BadConfig("Invalid channel specified.", config="channel")

    v = rgb[:, :, channel]
    return numpy.dstack((v, v, v))



import sys
import warnings

from pyparsing import lineno, col

from .block_meta import split_docstring, BlockInput, BlockOutput, BlockConfig
from .constants import FIXED
from .exceptions import SemanticError, x_not_found


class Where:
    """ An object of this class represents a place in a file. 
    
    All parsed elements contain a reference to a :py:class:`Where` object
    so that we can output pretty error messages.
    """

    def __init__(self, filename, string, character=None, line=None, column=None):
        self.filename = filename
        self.string = string
        if character is None:
            assert line is not None and column is not None
            self.line = line
            self.col = column
            self.character = None
        else:
            assert line is None and column is None
            self.character = character
            self.line = lineno(character, string)
            self.col = col(character, string)

    def __str__(self):
        s = ""
        s += "In file %s:\n" % self.filename
        context = 3
        lines = self.string.split("\n")
        start = max(0, self.line - context)
        pattern = "line %2d >"
        for i in range(start, self.line):
            s += "%s%s\n" % (pattern % (i + 1), lines[i])

        fill = len(pattern % (i + 1))
        space = " " * fill + " " * (self.col - 1)
        s += space + "^\n"
        s += space + "|\n"
        s += space + "here or nearby"
        return s

    def print_where(self, s=sys.stdout):
        # XXX: repeated code
        s.write("\n\n")
        prefix = "    "
        write = lambda x: s.write(prefix + x)
        write("In file %s:\n" % self.filename)
        context = 3
        lines = self.string.split("\n")
        start = max(0, self.line - context)
        pattern = "line %2d >"
        for i in range(start, self.line):
            write("%s%s\n" % (pattern % (i + 1), lines[i]))

        fill = len(pattern % (i + 1))
        space = " " * fill + " " * (self.col - 1)
        write(space + "^\n")
        write(space + "|\n")
        write(space + "here or nearby")


class ParsedElement:
    def __init__(self):
        self.where = None


class ParsedSignalList(ParsedElement):
    def __init__(self, signals):
        ParsedElement.__init__(self)
        self.signals = signals

    def __repr__(self):
        return "Signals%s" % self.signals

    @staticmethod
    def from_tokens(tokens):
        return ParsedSignalList(list(tokens))


class ImportStatement(ParsedElement):
    def __init__(self, package):
        ParsedElement.__init__(self)
        self.package = package

    def __repr__(self):
        return "import(%s)" % self.package

    @staticmethod
    def from_tokens(tokens):
        package = "".join(tokens.asList())
        return ImportStatement(package)


class ParsedSignal(ParsedElement):
    """ Note that the convention is tricky: :: 
    
             --> [local_input] name [local_output] -->
    """

    def __init__(self, name, block_name, local_input, local_output):
        ParsedElement.__init__(self)
        self.name = name
        self.block_name = block_name
        self.local_input = local_input
        self.local_output = local_output

    def __repr__(self):
        s = "Signal("
        if self.local_input is not None:
            s += "[%s]" % self.local_input
        if self.block_name is not None:
            s += "%s." % self.block_name
        s += str(self.name)
        if self.local_output is not None:
            s += "[%s]" % self.local_output
        s += ")"
        return s

    @staticmethod
    def from_tokens(tokens):
        name = tokens.get("name")
        block_name = tokens.get("block_name", None)
        local_input = tokens.get("local_input", None)
        local_output = tokens.get("local_output", None)
        return ParsedSignal(name, block_name, local_input, local_output)


class ParsedBlock(ParsedElement):
    def __init__(self, name, operation, config):
        ParsedElement.__init__(self)
        if operation in ["input", "output"]:
            if not isinstance(config.get("name"), str):
                raise ValueError("Invalid port name %r." % name)
        self.name = name
        self.operation = operation
        self.config = config

    def __repr__(self):
        return "Block(op=%s,name=%s,config=%s)" % (self.operation, self.name, self.config)

    @staticmethod
    def from_tokens(tokens):
        blocktype = tokens["blocktype"]
        config = tokens.get("config", {})
        name = tokens.get("name", None)
        return ParsedBlock(name, blocktype, config)


def parse_input_port(tokens):
    blocktype = "input"
    port_name = tokens.get("port_name")
    assert isinstance(port_name, str)
    config = dict(name=port_name)
    name = None
    return ParsedBlock(name, blocktype, config)


def parse_output_port(tokens):
    blocktype = "output"
    port_name = tokens.get("port_name")
    config = dict(name=port_name)
    name = None
    return ParsedBlock(name, blocktype, config)


class ParsedAssignment(ParsedElement):
    def __init__(self, key, value):
        ParsedElement.__init__(self)

        self.key = key
        self.value = value

    def __repr__(self):
        return "Assignment(%s=%s)" % (self.key, self.value)

    @staticmethod
    def from_tokens(tokens):
        return ParsedAssignment(tokens["key"], tokens["value"])


def config_from_tokens(tokens):
    variable = tokens.get("variable")
    has_default = "default" in tokens
    default = tokens.get("default", None)
    docstring = tokens.get("docstring", None)
    desc, desc_rest = split_docstring(docstring)
    # TODO: dtype for BlockConfig
    return BlockConfig(variable, has_default, default, desc, desc_rest, where=None, dtype=None)


def output_from_tokens(tokens):
    name = tokens.get("name")
    docstring = tokens.get("docstring", None)

    desc, desc_rest = split_docstring(docstring)

    return BlockOutput(FIXED, name, desc, desc_rest, None, dtype=None)


def input_from_tokens(tokens):
    name = tokens.get("name")
    docstring = tokens.get("docstring", None)

    desc, desc_rest = split_docstring(docstring)
    return BlockInput(FIXED, name, None, None, desc, desc_rest, None, dtype=None)


class Connection(ParsedElement):
    def __init__(self, elements):
        ParsedElement.__init__(self)
        self.elements = elements

    def __repr__(self):
        return "Connection(%s)" % self.elements

    @staticmethod
    def from_tokens(tokens):
        return Connection(tokens)


class VariableReference(ParsedElement):
    def __init__(self, variable):
        ParsedElement.__init__(self)
        self.variable = variable

    def __repr__(self):
        return "ref:%s" % self.variable

    @staticmethod
    def from_tokens(tokens):
        return VariableReference(tokens["variable"])


class ParsedModel(ParsedElement):
    # temporary storage while parsing
    static_filename = "not set"

    def __init__(self, name, docstring, elements):
        ParsedElement.__init__(self)
        assert name is None or isinstance(name, str)
        assert docstring is None or isinstance(docstring, str)
        self.name = name
        self.docstring = docstring

        select = lambda T: [x for x in elements if isinstance(x, T)]

        self.connections = select(Connection)
        self.config = select(BlockConfig)
        self.output = select(BlockOutput)
        self.input = select(BlockInput)
        self.imports = select(ImportStatement)
        self.assignments = select(ParsedAssignment)

        self.elements = elements

        # look for other input/output models
        def look_for_blocks(condition):
            for connection in self.connections:
                for element in connection.elements:
                    if condition(element):
                        yield element

        input_blocks = list(look_for_blocks(lambda x: isinstance(x, ParsedBlock) and x.operation == "input"))
        output_blocks = list(
            look_for_blocks(lambda x: isinstance(x, ParsedBlock) and x.operation == "output")
        )

        # print input_blocks
        # print output_blocks

        for block in input_blocks:
            inputs_defined = [x.name for x in self.input]
            input_name = block.config.get("name", None)
            if input_name is not None:
                # if a name is specified, check if some input
                # were specified formally
                # if some inputs were specified, it should be there
                # (we don't mix the two cases)
                if inputs_defined:
                    if not input_name in inputs_defined:
                        msg = x_not_found("input", input_name, inputs_defined)
                        raise SemanticError(msg, block)
                    else:
                        # good! this was already specified
                        pass
                else:
                    # we have a name, and no input was specified,
                    # so we add it (with warning)
                    bi = BlockInput(
                        type=FIXED,
                        name=input_name,
                        min=None,
                        max=None,
                        desc=None,
                        desc_rest=None,
                        where=block.where,
                        dtype=None,
                    )
                    # TODO: dtype for BlockInput
                    self.input.append(bi)
            else:
                # we don't have a name

                # if no input was defined then add it
                if not inputs_defined:
                    block.config["name"] = "in%d" % len(self.input)
                    bi = BlockInput(
                        type=FIXED,
                        name=block.config["name"],
                        min=None,
                        max=None,
                        desc=None,
                        desc_rest=None,
                        where=block.where,
                        dtype=None,
                    )
                    warnings.warn("check dtype")
                    # TODO add warning
                    self.input.append(bi)
                else:
                    # if exactly 1 input is specified, use that
                    if len(inputs_defined) == 1:
                        block.config["name"] = self.input[0].name
                    else:
                        # otherwise fail
                        msg = (
                            "This input block did not specify a name, and "
                            "I do not know which input it refers to."
                        )
                        raise SemanticError(msg, block)

        for block in output_blocks:
            outputs_defined = [x.name for x in self.output]
            output_name = block.config.get("name", None)
            if output_name is not None:
                # if a name is specified, check if some output
                # were specified formally
                # if some outputs were specified, it should be there
                if outputs_defined:
                    if not output_name in outputs_defined:
                        msg = x_not_found("output", output_name, outputs_defined)
                        raise SemanticError(msg, block)
                    else:
                        # good! this was already specified
                        pass
                else:
                    # we have a name, and no output was specified,
                    # so we add it (with warning)
                    bo = BlockOutput(
                        type=FIXED, name=output_name, desc=None, desc_rest=None, where=block.where, dtype=None
                    )
                    # TODO add warning
                    self.output.append(bo)
            else:
                # we don't have a name

                # if no output was defined then add it
                if not outputs_defined:
                    block.config["name"] = "out%d" % len(self.output)
                    bo = BlockOutput(
                        type=FIXED,
                        name=block.config["name"],
                        desc=None,
                        desc_rest=None,
                        where=block.where,
                        dtype=None,
                    )
                    self.output.append(bo)
                else:
                    # if exactly 1 output is specified, use that
                    if len(outputs_defined) == 1:
                        block.config["name"] = self.output[0].name
                    else:
                        # otherwise fail
                        msg = (
                            "This output block did not specify a name, and "
                            "I do not know which output it refers to."
                        )
                        raise SemanticError(msg, block)

    def __repr__(self):
        s = "Model %r: %d elements\n" % (self.name, len(self.elements))

        parts = [
            ("config parameters", self.config),
            ("input signals", self.input),
            ("output signals", self.output),
            ("import statements", self.imports),
            ("assignments", self.assignments),
            ("blocks connections", self.connections),
        ]

        for (part, elements) in parts:
            if len(elements) == 0:
                s += "- No %s.\n" % part
            else:
                s += "- %d %s:\n" % (len(elements), part)
                for e in elements:
                    s += "  * %r\n" % e
        return s

    @staticmethod
    def from_named_model(tokens):
        name = tokens["model_name"]
        elements = list(tokens["content"])
        docstring = tokens.get("docstring", None)

        return ParsedModel(name, docstring, elements)

    @staticmethod
    def from_anonymous_model(tokens):
        elements = list(tokens)
        return ParsedModel(name=None, docstring=None, elements=elements)



from .containers import CONTAINER_MOV, CONTAINER_MKV, CONTAINER_MP4
from contracts import contract
from procgraph.utils.calling_ext_program import system_cmd_result
from procgraph_mplayer.conversions.containers import CONTAINER_AVI

VCODEC_X264 = "x264"
VCODEC_PRORES = "prores"


@contract(profile="0|1|2|3")
def get_prores_encoder_params(profile=3, qv=None):
    assert profile in [0, 1, 2, 3]
    cmds = []
    cmds += ["-vcodec", "prores_kostya"]
    cmds += ["-profile:v", str(profile)]
    if qv is not None:
        cmds += ["-q:v", str(qv)]
    return cmds


def get_x264_encoder_params(crf=18, preset="medium"):
    cmds = []
    cmds += ["-vcodec", "libx264"]
    # Let's detect ffmpeg version
    res = system_cmd_result(".", ["ffmpeg", "-version"])
    ffmpeg_version = res.stdout.split("\n")[0]

    # SVN-r0.5.9-4:0.5.9-0ubuntu0.10.04.3

    # if 'ubuntu0.10.04.3' in ffmpeg_version:
    if "0.5" in ffmpeg_version:  # or '0.8.6' in ffmpeg_version:
        cmds += ["-vpre", "libx264-default"]
    else:
        cmds += ["-preset", preset]

    cmds += ["-crf", "%d" % crf]
    return cmds


VCODECS = {VCODEC_PRORES: get_prores_encoder_params, VCODEC_X264: get_x264_encoder_params}


@contract(returns="tuple(str, dict(str:*))")
def guess_vcodec(container):
    choices = {
        CONTAINER_MP4: (VCODEC_X264, dict()),
        CONTAINER_MKV: (VCODEC_X264, dict()),
        CONTAINER_MOV: (VCODEC_PRORES, dict()),
        CONTAINER_AVI: (VCODEC_X264, dict()),  # XXX: not sure
    }
    return choices[container]



from ..core.block import Block
from ..core.exceptions import SemanticError, PGSyntaxError, add_prefix
from ..core.model_loader import model_from_string
from ..core.parsing import parse_model
from ..core.registrar import default_library, Library

# noinspection PyUnresolvedReferences
import procgraph.components  # @UnusedImport
import unittest
import traceback


# load standard components


def define_generic(nin, nout):
    # noinspection PyUnusedLocal
    class Generic(Block):
        Block.alias("generic_in%d_out%d" % (nin, nout))
        for i in range(nin):
            Block.input(str(i))
        for i in range(nout):
            Block.output(str(i))


for nin in range(0, 6):
    for nout in range(0, 6):
        define_generic(nin, nout)


class PGTestCase(unittest.TestCase):
    def check_syntax_ok(self, model_spec):
        """ Tests that the given string can parse OK.
            Returns parsed models. """
        try:
            parsed = parse_model(model_spec)
            return parsed
        except Exception as e:
            traceback.print_exc()
            self.fail("Could not parse model:\n%s\nGot error: %s" % (self.write_fancy(model_spec), e))

    def check_semantic_ok(self, model_spec, config={}):
        """ Tests that the given string can parse OK and we can create a model.
             Note that a syntax error is translated into a test Error,
             not a failure.
        """
        # Don't pollute the main library with unit tests
        library = Library(parent=default_library)

        try:
            model = model_from_string(model_spec, config=config, library=library)
            model.init()
            return model
        except SemanticError as e:
            traceback.print_exc()

            self.fail(
                "Semantic error for not parse model:\n%s\n"
                "Got error: %s" % (self.write_fancy(model_spec), e)
            )
        except BaseException as e:
            self.fail(
                "Other error for not parse model:\n%s\n" "Got error: %s" % (self.write_fancy(model_spec), e)
            )

    def check_syntax_error(self, model_spec):
        """ Tests that the given string parsing gives a syntax error. """
        try:
            parsed = parse_model(model_spec)
            self.fail(
                "Expected SemanticError for model:\n%s\nParsed as:\n%s"
                % (self.write_fancy(model_spec), parsed)
            )
        except PGSyntaxError as e:
            e.__str__()
            return e

    def check_semantic_error(self, model_spec, config={}):
        """ Tests that the given string parses ok
            but gives a semantic error. """
        # make sure we can parse it, before
        parsed = parse_model(model_spec)

        try:
            library = Library(parent=default_library)
            model = model_from_string(model_spec, config=config, library=library)
            model.init()
            # Note: we only have check_semantic_error, perhaps we need
            # another for the tests that do need init()?
            self.fail("Expected SemanticError for model:\n%s\n%s" % (self.write_fancy(model_spec), parsed))

        except SemanticError as e:
            pass
            # make sure we can write it as a string
            e.__str__()
            return e

    def check_execution_throws(self, possible_errors, model_spec, config={}):
        """ Tests that the model execution throws the specific error
            semantic error. Returns the exception if it is thrown. """
        library = Library(parent=default_library)
        model = model_from_string(model_spec, config=config, library=library)
        try:
            model.init()
            while model.has_more():
                model.update()
            model.finish()
            self.fail(
                "Expected one of %s for model:\n%s" % (str(possible_errors), self.write_fancy(model_spec))
            )
        except possible_errors as e:
            # make sure we can write e as a string:
            str(e)
            return e

    def write_fancy(self, model_spec):
        return add_prefix(model_spec, "  |")


class VerifyBlock(Block):
    """
        This debug block verifies that config.x == config.y
        and throws an exception if that's not the case.
    """

    Block.alias("verify")

    Block.config("x")
    Block.config("y")

    def init(self):
        if self.config.x != self.config.y:
            raise SemanticError('Oops: "%s" != "%s".' % (self.config.x, self.config.y), self)



import numpy

from procgraph import Block, BadInput, register_model_spec


def isiterable(x):
    try:
        iter(x)
        return True
    except TypeError:
        return False


class ForwardDifference12(Block):
    """ Computes ``x[t+1] - x[t]`` normalized with timestamp. """

    Block.alias("two_step_difference")

    Block.input("x12", "An array with the last 2 values of x.")
    Block.input("t12", "An array with the last 2 values of the timestamp.")

    Block.output("x_dot", "Derivative of x")

    def update(self):
        x = self.input.x12
        t = self.input.t12
        if not isiterable(x) or len(x) != 2:
            raise BadInput("Expected arrays of 2 elements", self, "x")
        if not isiterable(t) or len(t) != 2:
            raise BadInput("Expected arrays of 2 elements", self, "t")

        delta = t[1] - t[0]

        if not delta > 0:
            raise BadInput("Bad timestamp sequence % s" % t, self, "t")

        # if this is a sequence of bytes, let's promove them to floats
        if x[0].dtype == numpy.dtype("uint8"):
            diff = x[1].astype("float32") - x[0].astype("float32")
        else:
            diff = x[1] - x[0]
        time = t[0]
        x_dot = diff / numpy.float32(delta)
        self.set_output("x_dot", x_dot, timestamp=time)


# TODO: move this to models/
register_model_spec(
    """
--- model derivative2
''' Computes the derivative of a quantity with 2 taps (``x[t+1] - x[t]``).
    See also :ref:`block:derivative`.       '''
input x "quantity to derive"
output x_dot "approximate derivative"

|input name=x| --> |last_n_samples n=2| --> x,t

   x, t --> |two_step_difference| --> |output name=x_dot|
    
"""
)



from .conversions import pg_video_info
from contracts.main import check
from procgraph import BadConfig, Block, Generator
from procgraph.block_utils import expand
from procgraph.utils import friendly_path
from procgraph_mplayer.programs_existence import check_programs_existence
import math
import numpy
import os
import subprocess
import tempfile


__all__ = ["MPlayer"]


class MPlayer(Generator):
    """ Decodes a video stream. """

    Block.alias("mplayer")

    Block.config("file", "Input video file. This can be in any format that " "``mplayer`` understands.")
    Block.config("quiet", "If true, suppress stderr messages from mplayer.", default=True)
    Block.config("stats", "If true, writes some statistics about the " "remaining time.", default=True)

    Block.config(
        "max_duration",
        "Maximum length, in seconds, of the output." "Useful to get a maximum duration.",
        default=None,
    )

    Block.output("video", "RGB stream as numpy array.")

    TIMESTAMPS_SUFFIX = ".timestamps"

    def init(self):
        self.programs = check_programs_existence(programs=["mencoder"])

        for p in self.programs:
            self.info("Using %13s = %s" % (p, self.programs[p]))

        if not isinstance(self.config.file, str):
            raise BadConfig("This should be a string.", self, "file")

        self.file = expand(self.config.file)

        if not os.path.exists(self.file):
            msg = "File %r does not exist." % self.file
            raise BadConfig(msg, self, "file")

        timestamps_file = self.file + MPlayer.TIMESTAMPS_SUFFIX
        if os.path.exists(timestamps_file):
            self.info("Reading timestamps from %r." % timestamps_file)
            self.timestamps = open(timestamps_file)
        else:
            self.timestamps = None
            # self.info('Will use fps for timestamps.')

        self.mencoder_started = False

        self.state.timestamp = None
        self.state.timestamp = self.get_next_timestamp()
        self.state.next_frame = None
        self.state.finished = False

        self.num_frames_read = 0

    def open_mencoder(self):
        self.mencoder_started = True

        info = pg_video_info(self.file, intolerant=True)

        self.width = info["width"]
        self.height = info["height"]

        self.fps = info["fps"]
        self.length = info["length"]

        check("float|int", self.length)
        check("float|int", self.fps)
        self.info("length: %r" % self.length)
        self.info("fps: %r" % self.fps)
        self.approx_frames = int(math.ceil(self.length * self.fps))

        # TODO: reading non-RGB streams not supported
        self.info(
            "Reading %dx%d @ %.3f fps "
            " (length %ss, approx %d frames), from %s."
            % (
                self.width,
                self.height,
                self.fps,
                self.length,
                self.approx_frames,
                friendly_path(self.config.file),
            )
        )

        self.shape = (self.height, self.width, 3)
        self.dtype = "uint8"

        pixel_format = "rgb24"

        self.temp_dir = tempfile.mkdtemp(prefix="procgraph_fifo_dir")
        self.fifo_name = os.path.join(self.temp_dir, "mencoder_fifo")
        os.mkfifo(self.fifo_name)
        args = [
            self.programs["mencoder"],
            self.file,
            "-ovc",
            "raw",
            "-rawvideo",
            "w=%d:h=%d:format=%s" % (self.width, self.height, pixel_format),
            "-of",
            "rawvideo",
            "-vf",
            "format=rgb24",
            "-nosound",
            "-o",
            self.fifo_name,
        ]

        self.tmp_stdout = tempfile.TemporaryFile()
        self.tmp_stderr = tempfile.TemporaryFile()

        if self.config.quiet:
            self.process = subprocess.Popen(
                args, stdout=self.tmp_stdout.fileno(), stderr=self.tmp_stderr.fileno()
            )
        else:
            self.process = subprocess.Popen(args)

        self.delta = 1.0 / self.fps

        if not self.timestamps:
            msg = "No timestamps found; using delta = %.2f (%.2f fps)." % (self.delta, self.fps)
            self.info(msg)

        self.stream = open(self.fifo_name, "r")

    def get_next_timestamp(self):
        if self.timestamps:
            # If reading from files
            l = self.timestamps.readline()
            if not l:
                # XXX warn if not even one
                self.error_once("Timestamps were too short; now using incremental.")
                if self.state.timestamp is not None:
                    return self.state.timestamp + self.delta
                else:
                    # empty file, not even one
                    self.error_once("Empty timestamp file? Starting at 0.")
                    return 0.0
            else:
                return float(l)
        else:
            if self.state.timestamp is None:
                return 0.0
            else:
                return self.state.timestamp + self.delta

    def update(self):
        if not self.mencoder_started:
            self.open_mencoder()
            self.read_next_frame()

        self.set_output(0, value=self.state.next_frame, timestamp=self.state.timestamp)
        self.state.timestamp = self.get_next_timestamp()

        self.state.next_frame = None
        self.read_next_frame()

        if self.config.stats:
            self.num_frames_read += 1
            if self.num_frames_read % 500 == 0:
                self.print_stats()

    def print_stats(self):
        if self.approx_frames != 0:
            percentage = 100.0 * self.num_frames_read / self.approx_frames
        else:
            percentage = 0
        # this assumes constant fps
        seconds = self.num_frames_read * self.delta
        seconds_total = self.approx_frames * self.delta
        self.info(
            "%6d/%d frames, %.1f/%.1f sec (%4.1f%%) of %s"
            % (
                self.num_frames_read,
                self.approx_frames,
                seconds,
                seconds_total,
                percentage,
                friendly_path(self.file),
            )
        )

    def read_next_frame(self):
        if self.state.finished:
            return
        if self.state.next_frame is not None:
            return

        # TODO: add display of stderr after timeout

        dtype = numpy.dtype(("uint8", self.shape))
        rgbs = numpy.fromfile(self.stream, dtype=dtype, count=1)

        if len(rgbs) == 0:
            self.state.next_frame = None
            self.state.finished = True
        else:
            self.state.next_frame = rgbs[0, :].squeeze()

    def too_long_for_us(self):
        """ Returns true if max_duration is set and we passed it. """
        if self.config.max_duration is None:
            return False

        # We haven't started yet, so we don't have "delta"
        if not self.mencoder_started:
            return False

        passed = self.num_frames_read * self.delta

        if passed > self.config.max_duration:
            self.info_once(
                "Finishing because passed %d frames = %f seconds > duration %f"
                % (self.num_frames_read, passed, self.config.max_duration)
            )
            return True
        else:
            return False

    def next_data_status(self):
        # The stream has finished
        if self.state.finished:
            return (False, None)
        # We passed the limit
        elif self.too_long_for_us():
            return (False, None)
        else:
            return (True, self.state.timestamp)

    def finish(self):
        # TODO: make sure process is closed?
        if os.path.exists(self.fifo_name):
            os.unlink(self.fifo_name)
        if os.path.exists(self.temp_dir):
            os.rmdir(self.temp_dir)


#
# # backported from 2.7
# def check_output(*popenargs, **kwargs):
#     """
#     Run command with arguments and return its output as a byte string.
#
#     If the exit code was non-zero it raises a CalledProcessError.  The
#     CalledProcessError object will have the return code in the returncode
#     attribute and output in the output attribute.
#
#     The arguments are the same as for the Popen constructor.  Example:
#
#     >>> check_output(["ls", "-l", "/dev/null"])
#     'crw-rw-rw- 1 root root 1, 3 Oct 18  2007 /dev/null\n'
#
#     The stdout argument is not allowed as it is used internally.
#     To capture standard error in the result, use stderr=STDOUT.
#
#     >>> check_output(["/bin/sh", "-c",
#     ...               "ls -l non_existent_file ; exit 0"],
#     ...              stderr=STDOUT)
#     'ls: non_existent_file: No such file or directory\n'
#     """
#     if 'stdout' in kwargs:
#         raise ValueError('stdout argument not allowed, it will be overridden.')
#     process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
#     output, unused_err = process.communicate()
#     retcode = process.poll()
#     if retcode:
#         cmd = kwargs.get("args")
#         if cmd is None:
#             cmd = popenargs[0]
#         raise subprocess.CalledProcessError(retcode, cmd, output=output)
#     return output



from procgraph import Block, ETERNITY


__all__ = ["Constant"]


class Constant(Block):
    """ Output a numerical constant that never changes.
    
        Example: ::
    
            |constant value=42| -> ...
            
    """

    Block.alias("constant")

    Block.config("value", "Constant value to output.")
    Block.output("constant", "The constant value.")

    def update(self):
        # XXX: are you sure we need ETERNITY?
        self.set_output(0, self.config.value, timestamp=ETERNITY)

    def __repr__(self):
        return "Constant(%s)" % self.config.value



from contracts import describe_type

import pickle

from .debug_pickler import find_pickling_error
from .safe_write import safe_write, safe_read


__all__ = ["safe_pickle_dump", "safe_pickle_load"]


def safe_pickle_dump(value, filename, protocol=pickle.HIGHEST_PROTOCOL, **safe_write_options):
    from procgraph import logger

    with safe_write(filename, **safe_write_options) as f:
        try:
            pickle.dump(value, f, protocol)
        except Exception:
            msg = "Cannot pickle object of class %s" % describe_type(value)
            logger.error(msg)
            msg = find_pickling_error(value, protocol)
            logger.error(msg)
            raise


def safe_pickle_load(filename):
    # TODO: add debug check
    with safe_read(filename) as f:
        return pickle.load(f)
        # TODO: add pickling debug



import warnings

from numpy.ma.testutils import assert_equal


__all__ = ["PlotGeneric"]


class PlotGeneric(object):
    """ 
        Produces images using matplotlib. Good for custom animation.
    """

    def __init__(self, width, height, transparent, tight, keep=False):
        self.width = width
        self.height = height
        self.transparent = transparent
        self.tight = tight
        self.keep = keep

        self.figure = None
        self.warned = False

    def init_figure(self):
        # TODO: remove from here
        from . import pylab

        pylab.rc("xtick", labelsize=8)
        pylab.rc("ytick", labelsize=8)

        """ Creates figure object and axes """
        self.figure = pylab.figure(frameon=False, figsize=(self.width / 100.0, self.height / 100.0))
        self.axes = pylab.axes()
        self.figure.add_axes(self.axes)

        pylab.draw_if_interactive = lambda: None

        pylab.figure(self.figure.number)

    def get_rgb(self, function):
        """ function(pylab) """
        if self.figure is None:
            self.init_figure()

        from . import pylab, pylab2rgb

        pylab.figure(self.figure.number)

        function(pylab)

        # http://matplotlib.sourceforge.net/users/tight_layout_guide.html
        try:
            pylab.tight_layout()
        except Exception as e:
            msg = "Could not call tight_layout(); available only on " "Matplotlib >=1.1 (%s)" % e
            if not self.warned:
                warnings.warn(msg)
                self.warned = True

        # There is a bug that makes the image smaller than desired
        # if tight is True
        pixel_data = pylab2rgb(transparent=self.transparent, tight=self.tight)

        from procgraph_images import image_pad  # need here otherwise circular

        # So we check and compensate
        shape = pixel_data.shape[0:2]
        shape_expected = (self.height, self.width)
        if shape != shape_expected:
            msg = "pylab2rgb() returned size %s instead of %s." % (shape, shape_expected)
            msg += " I will pad the image with white."
            warnings.warn(msg)
            pixel_data = image_pad(pixel_data, shape_expected, bgcolor=[1, 1, 1])
            assert_equal(pixel_data.shape[0:2], shape_expected)

        if not self.keep:
            pylab.close(self.figure.number)
            self.figure = None
        else:
            pass
        #             pylab.cla()
        #             pylab.clf()

        return pixel_data



from contracts import contract
import time


class InAWhile:
    @contract(interval="None|>0")
    def __init__(self, interval=5):
        self.interval = interval
        self.reset()

    def reset(self):
        self.count = 0
        self.start = time.time()
        self.last = self.start
        self.now = self.start
        pass

    def its_time(self):
        self.count += 1
        self.now = time.time()

        if self.interval is None:
            return False

        if self.now >= self.last + self.interval:
            self.last = self.now
            return True
        else:
            return False

    def fps(self):
        if self.count == 0:
            return 0
        return self.count / (self.now - self.start)



from .utils import PGTestCase

good_examples = [
    """
--- model ciao
config x1
config x2 = 2
config x3 = 2 "documentation"
config x4     "documentation"
config x5     '''Very long
documentation string'''

"""
]

bad_examples = [
    """
# cannot have newline before doc string
--- model ciao
config x3 = 2 
"documentation"
config x4     "documentation"

"""
]


class SyntaxTestMultiple(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from procgraph import Generator, Block, BadConfig
from procgraph.block_utils import expand
import os
import re

__all__ = ["FilesFromDir"]


class FilesFromDir(Generator):
    """
        This block reads the filenames from a directory according
        to a given regexp.

        For now the timestamp starts from 0 and it is fixed.
    """

    Block.alias("files_from_dir")
    Block.config("dir", "Directory containing the image files.")
    Block.config(
        "regexp",
        "Regular expression for images.",
        #                  default='(\w+)(\d+)\.(\w+)')
        default=".+\.(png|jpg)",
    )
    Block.config("fps", "Fixed frame per second.", default=20.0)
    Block.output("filename", "Image filename")

    def init(self):
        dirname = self.config.dir
        dirname = expand(dirname)

        if not os.path.exists(dirname):
            raise BadConfig("Not existent directory %r." % dirname, self, "dir")

        if not os.path.isdir(dirname):
            raise BadConfig("The file %r is not a directory." % dirname, self, "dir")

        regexp = self.config.regexp

        # TODO: use proper logging
        self.info("Reading %r from %r." % (regexp, dirname))
        all_files = os.listdir(dirname)

        selected = [os.path.join(dirname, f) for f in all_files if re.match(regexp, f)]

        def natural_key(string):
            """See http://www.codinghorror.com/blog/archives/001018.html"""
            return [int(s) if s.isdigit() else s for s in re.split(r"(\d+)", string)]

        selected.sort(key=lambda x: natural_key(os.path.basename(x)))

        self.info(f"Selected {len(selected)}/{len(all_files)} files.")

        fps = float(self.config.fps)
        if fps <= 0:
            raise BadConfig(self, f"Invalid fps value {fps!r}.", "fps")
        # tuples (timestamp, filename)
        frames = [(i / fps, f) for i, f in enumerate(selected)]

        if not frames:
            raise Exception("No frames found in dir %r." % dirname)

        self.state.frames = frames
        self.state.next_frame = 0

    def next_data_status(self):
        k = self.state.next_frame
        if k is None:
            return False, None
        else:
            frames = self.state.frames
            timestamp = frames[k][0]
            return True, timestamp

    def update(self):
        frames = self.state.frames
        k = self.state.next_frame

        assert k < len(frames)

        timestamp, filename = frames[k]

        self.set_output(0, value=filename, timestamp=timestamp)

        if k + 1 >= len(frames):
            self.state.next_frame = None
        else:
            self.state.next_frame = k + 1



import io

import numpy as np

from procgraph import Block, Generator, simple_block

__all__ = ["StaticImage", "imread"]


class StaticImage(Generator):
    Block.alias("static_image")
    Block.config("file", "Static image to read")
    Block.output("rgb", "Image rgb")

    done: bool

    def init(self):
        self.done = False

    def next_data_status(self):
        # print('Status')
        if self.done:
            # print ('Return False, none')
            return False, None
        else:
            # print ('Return True, none')
            return True, 0.0  # XXX: not sure

    def update(self):
        # print('update')
        image = imread(self.config.file)
        self.set_output("rgb", image, 0.0)
        self.done = True


@simple_block
def imread(filename: str):
    """
        Reads an image from a filename into a numpy array. This can have
        different dtypes according to whether it's RGB, grayscale, RGBA, etc.

        :param filename: Image filename.
        :type filename: string

        :return: image: The image as a numpy array.
        :rtype: image
    """
    from . import Image

    try:
        im = Image.open(filename)
    except Exception as e:
        msg = f'Could not open filename "{filename}".'
        raise ValueError(msg) from e

    data = np.array(im)

    return data


@simple_block
def jpg2rgb(image_data: bytes):
    """ Reads JPG bytes as RGB"""
    from . import Image

    im = Image.open(io.BytesIO(image_data))
    im = im.convert("RGB")
    data = np.array(im)
    # print filename, data.shape, data.dtype

    assert data.ndim == 3
    assert data.dtype == np.uint8

    return data


@simple_block
def imread_rgb(filename: str):
    """
        Reads an image from a file using PIL and converts to an RGB array.

        :param filename: Filename
        :type filename: str

        :return: A numpy array.
        :rtype: ``array[HxWx3](uint8)``

    """
    from . import Image

    try:
        im = Image.open(filename)
    except Exception as e:
        msg = f'Could not open filename "{filename}": {e}'
        raise ValueError(msg)

    im = im.convert("RGB")
    data = np.array(im)
    # print filename, data.shape, data.dtype

    assert data.ndim == 3
    assert data.dtype == np.uint8

    return data



from procgraph import Block
from procgraph.block_utils import check_rgb
from . import np


__all__ = ["Border", "rgb_pad", "image_border"]


class Border(Block):
    """ Adds a block around the input image. """

    Block.alias("border")

    Block.input("rgb", "Input image.")
    Block.output("rgb", "Image with borders added around.")
    Block.config("color", "border color (0-1 rgb)", default=[0, 0, 0])
    Block.config("width", default=1)
    Block.config("left", "pixel length for left border", default=None)
    Block.config("right", "pixel length for right border", default=None)
    Block.config("top", "pixel length for top border", default=None)
    Block.config("bottom", "pixel length for bottom border", default=None)

    def update(self):
        check_rgb(self, "rgb")

        def df(x):
            if x is None:
                return self.config.width
            else:
                return x

        # TODO: check color
        self.output.rgb = image_border(
            self.input.rgb,
            left=df(self.config.left),
            right=df(self.config.right),
            top=df(self.config.top),
            bottom=df(self.config.bottom),
            color=df(self.config.color),
        )


def rgb_pad(height, width, color):
    pad = np.zeros((height, width, 3), dtype="uint8")
    for i in range(3):
        pad[:, :, i] = color[i] * 255
    return pad


def image_border(rgb, left=0, right=0, top=0, bottom=0, color=[1, 1, 1]):
    orig_shape = rgb.shape

    if left > 0:
        # note: do this every time because it changes throughout
        height, width = rgb.shape[0:2]
        pad = rgb_pad(height, left, color)
        rgb = np.hstack((pad, rgb))

    if right > 0:
        height, width = rgb.shape[0:2]
        pad = rgb_pad(height, right, color)
        rgb = np.hstack((rgb, pad))

    if top > 0:
        height, width = rgb.shape[0:2]
        pad = rgb_pad(top, width, color)
        rgb = np.vstack((pad, rgb))
        assert rgb.shape[0] == height + top

    if bottom > 0:
        height, width = rgb.shape[0:2]
        pad = rgb_pad(bottom, width, color)
        rgb = np.vstack((rgb, pad))
        assert rgb.shape[0] == height + bottom

    assert rgb.shape[0] == orig_shape[0] + top + bottom
    assert rgb.shape[1] == orig_shape[1] + left + right

    return rgb



from .block_meta import BlockConfig
from .exceptions import aslist, SemanticError
from .visualization import semantic_warning


__all__ = ["resolve_config"]


def resolve_config(list_of_config, given_config, block, STRICT=True):
    """
        Checks that the configuration passed is coherent with the definition.
        Returns a dictionary with the values to use.
    
        If STRICT is True, throws an error if we pass some configuration
        that was not properly defined in list_of_config.
    
        list_of_config: list of BlockConfig instances
    """

    for c in list_of_config:
        assert isinstance(c, BlockConfig)

    # First we get the names of the required config variables.
    defined_variables = set([x.variable for x in list_of_config])
    required_variables = set([x.variable for x in list_of_config if not x.has_default])
    passed_variables = set(given_config.keys())
    required_not_passed = required_variables.difference(passed_variables)
    passed_not_defined = passed_variables.difference(defined_variables)

    if required_not_passed:
        msg = "Some required config (%r) was not passed." % aslist(required_not_passed)
        raise SemanticError(msg, block)

    if passed_not_defined:
        msg = (
            "We were passed some config variable(s) (%r) which was not "
            "defined formally "
            "as a config variable.\nThe defined ones are: %s."
            % (aslist(passed_not_defined), aslist(defined_variables))
        )
        if STRICT:
            raise SemanticError(msg, block)
        else:
            semantic_warning(msg, block)

    result = {}
    for x in list_of_config:
        if x.has_default:
            result[x.variable] = x.default
    for key, val in given_config.items():
        result[key] = val

    return result



#!/usr/bin/env python
from optparse import OptionParser
from procgraph import logger
from procgraph.utils import wrap_script_entry_point
from procgraph_mplayer.conversions.video_info import pg_video_info
from pprint import pformat


usage = """
    %cmd  [-o <output>] <filename>
"""


def video_info_main(args):
    parser = OptionParser(usage=usage)

    #     parser.add_option("-o", dest='output', type="string",
    #                       help='Output video.')

    (_, args) = parser.parse_args(args)

    if len(args) != 1:
        msg = "Exactly one argument necessary."
        raise Exception(msg)

    filename = args[0]

    info = pg_video_info(filename)

    logger.info(info=info)

    return 0


def main():
    wrap_script_entry_point(video_info_main, logger)


if __name__ == "__main__":
    main()



import numpy as np


__all__ = ["Image_from_array"]


def Image_from_array(a):
    """ Converts an image in a numpy array to an Image instance.
        Accepts:  h x w      255  interpreted as grayscale
        Accepts:  h x w x 3  255  rgb  
        Accepts:  h x w x 4  255  rgba """

    # require_array(a)

    if not a.dtype == "uint8":
        raise ValueError('I expect dtype to be uint8, got "%s".' % a.dtype)

    if len(a.shape) == 2:
        height, width = a.shape
        rgba = np.zeros((height, width, 4), dtype="uint8")
        rgba[:, :, 0] = a
        rgba[:, :, 1] = a
        rgba[:, :, 2] = a
        rgba[:, :, 3] = 255
    elif len(a.shape) == 3:
        height, width = a.shape[0:2]
        depth = a.shape[2]
        rgba = np.zeros((height, width, 4), dtype="uint8")
        if not depth in [3, 4]:
            raise ValueError('Unexpected shape "%s".' % str(a.shape))
        rgba[:, :, 0:depth] = a[:, :, 0:depth]
        if depth == 3:
            rgba[:, :, 3] = 255
    else:
        raise ValueError('Unexpected shape "%s".' % str(a.shape))

    # require_shape((gt(0), gt(0), 4), rgba)
    from . import Image

    im = Image.frombuffer("RGBA", (width, height), rgba.data, "raw", "RGBA", 0, 1)
    return im



from procgraph import Block
from procgraph.utils import safe_pickle_dump, safe_pickle_load
import os

__all__ = ["CovarianceRemember"]


class CovarianceRemember(Block):
    """ Quick hack to remember covariance across executions """

    Block.alias("covariance_rem")

    Block.config("filename", default="cov_remember.pickle")
    Block.input("x")
    Block.output("cov_x")

    def init(self):
        filename = self.config.filename
        if os.path.exists(filename):
            self.info("Loading state from filename %r" % filename)
            self.state = safe_pickle_load(filename)
        else:
            from astatsa.mean_covariance import MeanCovariance

            self.state = MeanCovariance()

    def update(self):
        x = self.input.x.astype("float32")
        self.state.update(x)

        self.output.cov_x = self.state.get_covariance()
        safe_pickle_dump(self.state, self.config.filename)  # @UndefinedVariable



from procgraph import simple_block
from contracts import contract
import numpy as np


@simple_block
@contract(image="array[HxWx(3|4)]", bg="array[HxWx(3|4)]")
def bg_subtract(image, bg, perc=99, weighted=False):
    image_rbg = image[:, :, 0:3]
    bg_rgb = bg[:, :, 0:3]
    # pointwise difference
    error = image_diff(image_rbg, bg_rgb)

    threshold = np.percentile(error, perc)
    mask_robot = error > threshold
    mask_env = np.logical_not(mask_robot)

    H, W = image.shape[0:2]
    rgba = np.zeros((H, W, 4), "uint8")
    rgba[:, :, 0] = image[:, :, 0]
    rgba[:, :, 1] = image[:, :, 1]
    rgba[:, :, 2] = image[:, :, 2]

    if weighted:
        error -= threshold
        min_alpha = 100
        alpha = min_alpha + (255 - min_alpha) * (error / error.max())
        alpha[mask_env] = 0
        rgba[..., 3] = alpha
    else:
        rgba[mask_env, 3] = 0
        rgba[mask_robot, 3] = 255

    return rgba


def image_diff(a, b):
    return np.mean(np.abs(a.astype("float32") - b.astype("float32")), axis=2)



from . import json

from procgraph import Block


class AsJSON(Block):
    """ Converts the input into a JSON string. 
    
        TODO: add example
    """

    Block.alias("as_json")

    Block.input_is_variable("Inputs to transcribe as JSON.")

    Block.output("json", "JSON string.")

    def update(self):
        data = {}
        data["timestamp"] = max(self.get_input_signals_timestamps())
        for i in range(self.num_input_signals()):
            name = self.canonicalize_input(i)
            value = self.input[name]
            data[name] = value

        self.output.json = json.dumps(data)



from procgraph import Block


class HistoryN(Block):
    """ 
    This block collects the last N samples of a signals,
    and outputs two signals ``x`` and ``t``. 
    See also :ref:`block:historyt` and :ref:`block:history`.
    """

    Block.alias("last_n_samples")

    Block.config("n", "Number of samples to retain.")

    Block.input("x", "Any data")

    Block.output("x", "Sequence of values.")
    Block.output("t", "Sequence of timestamps.")

    def init(self):
        self.state.x = []
        self.state.t = []

    def update(self):

        sample = self.get_input(0)
        timestamp = self.get_input_timestamp(0)

        x = self.state.x
        t = self.state.t

        x.append(sample)
        t.append(timestamp)

        n = self.config.n
        while len(x) > n:
            t.pop(0)
            x.pop(0)

        if len(x) == n:
            self.output.x = x
            self.output.t = t



import os
import subprocess

import numpy

from procgraph import Block, BadConfig, ETERNITY
from procgraph.core.visualization import info as info_main, error as error_main
from typing import Optional, Tuple

from .pil_conversions import Image_from_array
from contracts import contract
from procgraph_pil.pil_format_paragraph import format_paragraph


__all__ = ["Text"]


class Text(Block):
    """ This block provides text overlays over an image.

    This block is very powerful, but the configuration is a bit complicated.

    You should provide a list of dictionary in the configuration variable
    ``texts``. Each dictionary in the list describes how and where to write
    one piece of text.

    An example of valid configuration is the following: ::

        text.texts = [{string: "raw image", position: [10,30], halign: left,
                      color: black, bg: white }]

    The meaning of the fields is as follow:

    ``string``
      Text to display. See the section below about keyword expansion.

    ``position``
      Array of two integers giving the position of the text in the image

    ``color``
      Text color. It can be a keyword color or an hexadecimal string
      (``white`` or ``#ffffff``).

    ``bg``
      background color

    ``halign``
      Horizontal alignment.
      Choose between ``left`` (default), ``center``, ``right``.

    ``valign``
      Vertical alignment.
      Choose between ``top`` (default), ``middle``, ``center``.

    ``size``
      Font size in pixels

    ``font``
      Font family. Must be a ttf file (``Arial.ttf``)

    **Expansion**: Also we expand macros in the text using ``format()``.
    The available keywords are:

    ``frame``
      number of frames since the beginning

    ``time``
      seconds since the beginning of the log

    ``timestamp``
      absolute timestamp

    """

    Block.alias("text")

    Block.config("texts", "Text specification (see block description).")

    Block.input("rgb", "Input image.")
    Block.output("rgb", "Output image with overlaid text.")

    def init(self):
        self.state.first_timestamp = None

    def update(self):

        # TODO: add check
        if self.state.first_timestamp is None:
            self.state.first_timestamp = self.get_input_timestamp(0)
            self.state.frame = 0
        else:
            self.state.frame += 1
        # Add stats
        macros = {}
        macros["timestamp"] = self.get_input_timestamp(0)
        if self.state.first_timestamp == ETERNITY:
            macros["time"] = -1
        else:
            macros["time"] = self.get_input_timestamp(0) - self.state.first_timestamp
        macros["frame"] = self.state.frame

        rgb = self.input.rgb
        im = Image_from_array(rgb)
        from . import ImageDraw

        draw = ImageDraw.Draw(im)

        # {string: "raw image", position: [10,30], halign: left,
        # color: black, bg: white  }
        if not isinstance(self.config.texts, list):
            raise BadConfig("Expected list", self, "texts")

        for text in self.config.texts:
            text = text.copy()
            if not "string" in text:
                raise BadConfig('Missing field "string" in text spec %s.' % text.__repr__(), self, "texts")

            try:
                text["string"] = Text.replace(text["string"], macros)
            except KeyError as e:
                msg = str(e) + "\nAvailable: %s." % list(macros.keys())
                raise BadConfig(msg, self, "texts")

            p = text["position"]

            try:
                p[0] = get_ver_pos_value(p[0], height=rgb.shape[0])
                p[1] = get_hor_pos_value(p[1], width=rgb.shape[1])
            except ValueError as e:
                raise BadConfig("%s: %s" % (e, text), self, "texts")

            process_text(draw, text)

        out = im.convert("RGB")
        pixel_data = numpy.asarray(out)

        self.output.rgb = pixel_data

    @staticmethod
    def replace(s, macros):
        """ Expand macros in the text. """
        return s.format(**macros)


def get_hor_pos_value(spec, width: int) -> int:
    if isinstance(spec, str):
        if is_fraction(spec):
            return get_fraction(spec, width)
        else:
            valid = ["center", "left", "right"]
            if not spec in valid:
                msg = "Strange %r: not in %s" % (spec, valid)
                raise ValueError(msg)
    else:
        assert isinstance(spec, int)

    if spec == "center":
        return int(width / 2)
    elif spec == "left":
        return 0
    elif spec == "right":
        return width
    else:
        if spec < 0:
            return width + spec
        else:
            return spec


def is_fraction(spec):
    return isinstance(spec, str) and spec[-1] == "%"


def get_fraction(spec, L):
    if not is_fraction(spec):
        raise ValueError(spec)
    f = float(spec[:-1])
    return int(f / 100.0 * L)


def get_ver_pos_value(spec, height):
    # Draws the string at the given position. The position gives the upper left corner of the text.
    if isinstance(spec, str):
        if is_fraction(spec):
            return get_fraction(spec, height)
        else:
            valid = ["middle", "top", "bottom"]
            if not spec in valid:
                msg = "Strange %r: not in %s" % (spec, valid)
                raise ValueError(msg)
    else:
        assert isinstance(spec, int)

    if spec == "middle":
        return int(height / 2)
    elif spec == "top":
        return 0
    elif spec == "bottom":
        return height
    else:
        if spec < 0:
            return height + spec
        else:
            return spec


def info(s):
    info_main("procgraph_pil/text: %s" % s)


def error(s):
    error_main("procgraph_pil/text: %s" % s)


# cache of fonts
def find_file(font_name: str) -> Optional[str]:
    try:
        # pattern = '*%s*.ttf' % font_name
        pattern = font_name
        a = subprocess.Popen(["locate", pattern], stdout=subprocess.PIPE)
        lines = a.stdout.read().decode()
        options = lines.split("\n")
        options = [f for f in options if os.path.exists(f)]
        if len(options) == 0:
            error("Cannot find a file matching the pattern %r." % pattern)
            return None
        guess = options[0]
        info('Found %d matches for %s, using  "%s".' % (len(options), pattern, guess))
        return guess
    except Exception as e:
        error('Cannot run "locate": %s' % e)
        return None


fonts = {}


def get_font(name, size):

    key = (name, size)
    # cached
    if key in fonts:
        return fonts[key]

    options = [name, "FreeSans", "Arial"]

    errors = []
    for o in options:
        try:
            res = get_font_try(o, size)
            fonts[key] = res
            return res
        except KeyError as e:
            errors.append(e)

    msg = "Could not find any of the fonts %r" % options
    raise ValueError(msg)


def get_font_try(name, size):
    from . import ImageFont

    name = name.replace(" ", "")
    filename = name
    if not filename.endswith(".ttf"):
        filename += ".ttf"
    if os.path.exists(filename):
        return ImageFont.truetype(filename, size)
    else:
        found = find_file(filename)
        if found is None:
            msg = "Could not find %r anywhere." % filename
            raise KeyError(msg)
        info('Found %r using "locate".' % filename)
        return ImageFont.truetype(found, size)


def process_text(draw, t):
    position = t["position"]
    string = t["string"]
    color = t.get("color", "#aaaaaa")
    bg = t.get("bg", None)
    size = t.get("size", 15)
    fontname = t.get("font", "Arial")
    font = get_font(fontname, size)

    tw, th = font.getsize(string)
    y, x = position[0], position[1]  # order is good

    halign = t.get("halign", "left")
    valign = t.get("valign", "top")

    if halign == "left":
        pass
    elif halign == "right":
        x -= int(tw)
    elif halign == "center":
        x -= int(tw / 2)
    else:
        print(("Unknown horizontal-align key %s" % halign))

    if valign == "top":
        pass
    elif valign == "bottom":
        y -= int(th)
    elif valign == "middle":
        y -= int(th / 2)
    else:
        print(f"Unknown vertical-align key {valign}")

    if bg:
        for a in [[-1, 0], [1, 0], [0, 1], [0, -1], [-1, -1], [-1, +1], [1, 1], [1, -1]]:
            draw.text([x + a[0], y + a[1]], string, fill=bg, font=font)

    draw.text([x, y], string, fill=color, font=font)


@contract(rgb="array[HxWx4]")
def draw_text_on_img(
    rgb,
    string,
    position,
    color="#aaaaaaff",
    bg=None,
    fontsize=None,
    height=None,
    width=None,
    fontname="Arial",
    line_width="100%",
    line_scale=1.5,
    halign="left",
    valign="top",
):

    im = Image_from_array(rgb)
    from . import ImageDraw

    draw = ImageDraw.Draw(im)

    position[0] = get_ver_pos_value(position[0], height=rgb.shape[0])
    position[1] = get_hor_pos_value(position[1], width=rgb.shape[1])

    print(f"Using line width {line_width!r} (shape {rgb.shape!r}) ")

    if is_fraction(line_width):
        line_width = get_fraction(line_width, rgb.shape[1])

    n = sum([fontsize is not None, height is not None, width is not None])
    if n != 1:
        msg = "Need exactly one of fontsize, height, width (got %d)" % n
        raise ValueError(msg)

    if fontsize is not None:
        if is_fraction(fontsize):
            fontsize = get_fraction(fontsize, rgb.shape[0])
        font = get_font(fontname, fontsize)
    else:
        test_scale = 100
        font0 = get_font(fontname, test_scale)
        _, (th0, tw0) = format_paragraph(text=string, font=font0, line_width=100000, line_scale=line_scale)

        print(f"size using test_scale ={test_scale} -> h {th0} w {tw0}")

        if height is not None:
            if is_fraction(height):
                height = get_fraction(height, rgb.shape[0])
            # print('height: %s' % height)
            scale = height * 1.0 / th0

        if width is not None:
            if is_fraction(width):
                width = get_fraction(width, rgb.shape[1])

            # print('width: %s' % width)
            scale = width * 1.0 / tw0

        use_size = int(test_scale * scale)
        print(("scale: %s" % scale))
        print(("use_size: %s" % use_size))
        font = get_font(fontname, use_size)

        _, (th1, tw1) = format_paragraph(text=string, font=font, line_width=100000, line_scale=line_scale)
        print(("now: h %s w %s " % (th1, tw1)))

    # print('font: %s' % str(font))
    tokens, (th, tw) = format_paragraph(text=string, font=font, line_width=line_width, line_scale=line_scale)
    print(("th, tw: %s" % str((th, tw))))
    base = get_basepoint(p0=position, tw=tw, th=th, halign=halign, valign=valign)
    print(("basepoint: %s" % str(base)))
    for t in tokens:
        # print(t)
        draw_token(draw=draw, t=t, base=base, bg=bg, font=font, color=color)

    out = im.convert("RGBA")
    pixel_data = numpy.asarray(out)
    return pixel_data


def draw_token(draw, t, base, bg, font, color):

    x = t.position[1] + base[1]
    y = t.position[0] + base[0]

    # print('drawing at y=%.2f x=%.2f token %r' % (y,x,t.text))
    string = t.text
    if bg:
        for a in [[-1, 0], [1, 0], [0, 1], [0, -1], [-1, -1], [-1, +1], [1, 1], [1, -1]]:
            draw.text([x + a[0], y + a[1]], string, fill=bg, font=font)

    # w, h = font.getsize(string)
    #     draw.rectangle(((0,0),(w,h)), fill="white", outline = "blue")
    # print('token %r -> size h %s w %s ' % (t, h ,w ))
    # draw.text([0, 0], string, fill="#000000", font=font)

    draw.text([x, y], string, fill=color, font=font)


def get_basepoint(p0: Tuple[int, int], tw, th, halign, valign) -> Tuple[int, int]:
    y, x = p0
    if halign == "left":
        pass
    elif halign == "right":
        x -= int(tw)
    elif halign == "center":
        x -= int(tw / 2)
    else:
        print(f"Unknown horizontal-align key {halign}")

    if valign == "top":
        pass
    elif valign == "bottom":
        y -= int(th)
    elif valign == "middle":
        y -= int(th / 2)
    else:
        print(f"Unknown vertical-align key {valign}")
    return y, x



""" Blocks for image operations based on the PIL library """


procgraph_info = {"requires": ["PIL"]}

# Smart dependencies initialization
from procgraph import import_magic, import_succesful

Image = import_magic(__name__, "PIL.Image")
ImageFont = import_magic(__name__, "PIL.ImageFont")
ImageDraw = import_magic(__name__, "PIL.ImageDraw")


from .pil_operations import *
from .pil_conversions import *
from .text import *
from .imread_imp import *
from .imwrite_imp import *



#!/usr/bin/env python
from optparse import OptionParser
from procgraph import pg, register_model_spec
import numpy as np
import os

usage = """
    %cmd  [-o <output>] <filename>
"""


def main():
    parser = OptionParser(usage=usage)

    parser.add_option("-o", dest="output", type="string", help="Output video.")

    (options, args) = parser.parse_args()

    if len(args) != 1:
        msg = "Exactly one argument necessary."
        raise Exception(msg)

    filename = args[0]

    return video_crop(filename, options.output)


def video_crop(filename, output=None):
    if output is None:
        base, ext = os.path.splitext(filename)
        output = "%s-crop%s" % (base, ext)

    mask = get_motion_mask(filename)
    # print mask
    H, W = mask.shape
    useless_line = [np.min(mask[i, :]) for i in range(H)]
    useless_col = [np.min(mask[:, i]) for i in range(W)]

    useful_line = np.logical_not(useless_line)
    useful_col = np.logical_not(useless_col)
    # print useful_line
    # print useful_col

    crop_top = np.min(np.nonzero(useful_line)[0])
    crop_bottom = H - np.max(np.nonzero(useful_line)[0])
    crop_left = np.min(np.nonzero(useful_col)[0])
    crop_right = W - np.max(np.nonzero(useful_col)[0])

    print(("Cropping left %s  right %s  top %s  bottom %s" % (crop_left, crop_right, crop_top, crop_bottom)))

    register_model_spec(
        """
--- model crop_movie
config video 'video'
config output
config top = 0
config left = 0
config right = 0
config bottom = 0

|mplayer file=$video stats=1| --> rgb

rgb --> |crop left=$left top=$top bottom=$bottom right=$right| --> rgb2

rgb2 --> |mencoder quiet=1 file=$output timestamps=0|

    """
    )

    pg(
        "crop_movie",
        dict(
            top=crop_top, bottom=crop_bottom, left=crop_left, right=crop_right, video=filename, output=output
        ),
    )


register_model_spec(
    """
--- model movie_stats
config video 'video'
output rgb_min
output rgb_max

|mplayer file=$video stats=1| --> rgb

rgb --> |minimum_over_time| --> |output name=rgb_min|
rgb --> |maximum_over_time| --> |output name=rgb_max|

    """
)


def get_motion_mask(filename):

    model = pg("movie_stats", dict(video=filename))
    rgb_min = model.get_output("rgb_min")
    rgb_max = model.get_output("rgb_max")

    no_motion = np.min(rgb_min == rgb_max, axis=2)

    def percentage(x):
        return 100.0 * np.sum(x.flat) / x.size

    print(("Immobile: %.1f%% of the image." % percentage(no_motion)))

    bgcolor = rgb_min[0, 0, :]

    print(("bgcolor: %s" % bgcolor))

    whiter = rgb_min[:, :, 0] == bgcolor[0]
    whiteg = rgb_min[:, :, 1] == bgcolor[1]
    whiteb = rgb_min[:, :, 2] == bgcolor[2]

    white = np.logical_and(np.logical_and(whiter, whiteg), whiteb)

    print(("White: %.1f%% of the image." % percentage(white)))

    mask = np.logical_and(white, no_motion)

    print(("Both: %.1f%% of the image." % percentage(mask)))

    return mask


if __name__ == "__main__":
    main()



""" 
    Blocks performing operations with a dynamic nature. 

    This library contains blocks that perform operations with time.
     
"""

from .derivative import *
from .derivative2 import *
from .history import *
from .sync import *
from .sieve import *
from .wait import *

from .historyt import *
from .fps_data_limit import *
from .fps_print import *
from .low_pass import *
from .last_n_samples import *
from .fps_limit import *

# static TODO: move?
from .join import *
from .extract import *
from .make_tuple import *

from .any import *

from .async_ import *
from .time_slice import *

from .all_ready import *
from .wait_sec import *
from .retime import *



# Autogenerated file -- do not modify.

# This file is generated by "pgindex".
# It lists the default ProcGraph packages and their dependencies
# so that we can warn the user if some dependency is not found at
# install time. ProcGraph's policy is to be very liberal; for example,
# the `procgraph_cv' package is installed even though OpenCV is not found.
# and an error will be thrown only if the user actually uses those blocks.

index = {
    "packages": {
        "procgraph.components.debug_components": {
            "blocks": ["info", "constant", "+", "*", "-", "/", "gain", "print", "identity"],
            "desc": "Components used for debugging and unit tests.",
            "requires": {},
        },
        "procgraph_cv": {
            "blocks": ["gradient", "smooth"],
            "desc": "Blocks using the OpenCV library.",
            "requires": {"cv": ["cv", "opencv"]},
        },
        "procgraph_foo": {
            "blocks": ["block_example"],
            "desc": "An example package for ProcGraph that shows how to organize your code. ",
            "requires": {"pickle": ["pickle"]},
        },
        "procgraph_hdf": {
            "blocks": ["hdfread_many", "hdfread_test", "hdfread_many_test", "hdfwrite", "hdfread"],
            "desc": "This is a set of blocks to read and write logs in HDF_ format. ",
            "requires": {"tables": ["tables"]},
        },
        "procgraph_images": {
            "blocks": [
                "rgb2gray",
                "compose",
                "reshape2d",
                "solid",
                "blend",
                "skim_top_and_bottom",
                "posterize",
                "gray2rgb",
                "scale",
                "grid",
                "skim_top",
                "posneg",
                "border",
            ],
            "desc": "Blocks for basic operations on images. ",
            "requires": {},
        },
        "procgraph_io_misc": {
            "blocks": ["pickle_load", "to_file", "pickle", "pickle_group", "as_json"],
            "desc": "Miscellaneous functions to be better organized.",
            "requires": {"json": ["simplejson"]},
        },
        "procgraph_mpl": {
            "blocks": ["plot"],
            "desc": "Blocks using Matplotlib to display data.",
            "requires": {"matplotlib": ["matplotlib"], "matplotlib.pylab": ["matplotlib.pylab"]},
        },
        "procgraph_mplayer": {
            "blocks": ["mplayer", "mencoder"],
            "desc": "Blocks for encoding/decoding video based on MPlayer.",
            "requires": {},
        },
        "procgraph_numpy_ops": {
            "blocks": [
                "smooth1d",
                "square",
                "sign",
                "minimum",
                "select",
                "rad2deg",
                "outer",
                "log",
                "sum",
                "astype",
                "arctan",
                "abs",
                "take",
                "real",
                "deg2rad",
                "hstack",
                "flipud",
                "max",
                "vstack",
                "gradient1d",
                "dstack",
                "maximum",
                "fliplr",
                "normalize_Linf",
                "mean",
            ],
            "desc": "Various operations wrapping numpy functions.",
            "requires": {},
        },
        "procgraph_pil": {
            "blocks": ["imread", "text", "resize"],
            "desc": "Blocks for image operations based on the PIL library",
            "requires": {"PIL": ["PIL"]},
        },
        "procgraph_robotics": {
            "blocks": [
                "pose2commands",
                "laser_display",
                "skim",
                "organic_scale",
                "pose2vel_",
                "laser_dot_display",
            ],
            "desc": "Some functions specific to robotics applications. ",
            "requires": {"geometry": ["geometry"]},
        },
        "procgraph_ros": {
            "blocks": ["bagread_test", "ros2python", "bagread"],
            "desc": "This is a set of blocks to read and write logs in ROS_ Bag format. ",
            "requires": {"ros": ["ros"]},
        },
        "procgraph_signals": {
            "blocks": [
                "derivative2",
                "slice",
                "two_step_difference",
                "derivative",
                "join",
                "sync",
                "fps_data_limit",
                "history",
                "historyt",
                "forward_difference",
                "low_pass",
                "make_tuple",
                "fps_limit",
                "fps_print",
                "last_n_samples",
                "sieve",
                "extract",
                "any",
                "wait",
            ],
            "desc": "Blocks performing operations with a dynamic nature. ",
            "requires": {},
        },
        "procgraph_statistics": {
            "blocks": ["cov2corr", "normalize", "expectation", "covariance", "soft_variance", "variance"],
            "desc": "Blocks for common statistical operations.",
            "requires": {},
        },
        "procgraph_yaml": {
            "blocks": ["yaml2object"],
            "desc": "YAML conversions.",
            "requires": {"yaml": ["yaml"]},
        },
    }
}



from procgraph import Block

__all__ = ["Identity"]


class Identity(Block):
    """ 
        This block outputs the inputs, unchanged. 
    """

    Block.alias("identity")
    Block.input_is_variable("Input signals.", min=1)
    Block.output_is_variable("Output signals, equal to input.")

    def update(self):
        # Just copy the input to the output
        for i in range(self.num_input_signals()):
            self.set_output(i, self.get_input(i), self.get_input_timestamp(i))



from numpy import array, cos, linspace, logical_not, maximum, minimum, nonzero, sin
from procgraph import BadInput, Block


__all__ = ["LaserDisplay"]


class LaserDisplay(Block):
    """ Produces a plot of a range-finder scan.

    Example of configuration: ::

        display_sick.groups = [{ indices: [0,179], theta: [-1.57,+1.57],
             color: 'r', origin: [0,0,0]}]

    """

    Block.alias("laser_display")

    Block.config("width", "Width of the resulting image.", default=320)
    Block.config("height", "Height of the resulting image.", default=320)
    Block.config("max_readings", "Readings are clipped at this threshold (m).", default=30)
    Block.config("groups", "How to group and draw the readings. (see example)")
    Block.config(
        "title", "By default it displays the signal name." " Set the empty string to disable.", default=None
    )

    Block.config("transparent", "Gives transparent RGBA rather than RGB.", default=False)
    Block.input("readings", "The laser readings (array of floats).")

    Block.output("image", "The laser visualization (rgba).")

    def update(self):

        readings = array(self.input.readings)

        if len(readings.shape) > 1:
            msg = "Expected flat array, got shape %s" % str(readings.shape)
            raise BadInput(msg, self, 0)

        from procgraph_mpl import pylab, pylab2rgb

        f = pylab.figure(frameon=False, figsize=(self.config.width / 100.0, self.config.height / 100.0))

        # limit the readings

        bounds = array([0, 0, 0, 0])

        for group in self.config.groups:
            indices = group["indices"]
            indices = list(range(indices[0], indices[-1] + 1))
            theta_spec = group["theta"]
            origin = group.get("origin", [0, 0, 0])
            color = group.get("color", "b.")

            max_readings = group.get("max_readings", self.config.max_readings)
            group_readings = minimum(readings[indices], max_readings)

            N = len(indices)

            theta = linspace(theta_spec[0], theta_spec[-1], N)

            assert len(theta) == len(group_readings)

            x = cos(theta + origin[2]) * group_readings + origin[0]
            y = sin(theta + origin[2]) * group_readings + origin[1]

            valid_flag = group_readings < max_readings
            (valid,) = nonzero(valid_flag)
            (invalid,) = nonzero(logical_not(valid_flag))

            pylab.plot(-y[valid], x[valid], color)
            pylab.plot(-y[invalid], x[invalid], "r.")

            R = max_readings * 1.1
            x_R = R * cos(theta)
            y_R = R * sin(theta)
            group_bounds = array([min(x_R), max(x_R), min(y_R), max(y_R)])
            for i in [1, 3]:
                bounds[i] = maximum(bounds[i], group_bounds[i])
            for i in [0, 2]:
                bounds[i] = minimum(bounds[i], group_bounds[i])

        pylab.axis(bounds)

        if self.config.title is not None:
            if self.config.title != "":
                pylab.title(self.config.title, fontsize=10)
        else:
            # We don't have a title ---
            t = self.get_input_signals_names()[0]
            pylab.title(t, fontsize=10)

        self.output.image = pylab2rgb(transparent=self.config.transparent)
        # tight=True)

        pylab.close(f.number)



from procgraph import Generator, Block
import numpy as np
from procgraph_mpl.plot_generic import PlotGeneric
from procgraph_mpl.plot_anim import PlotAnim


class JitteryClock(Generator):
    Block.alias("jittery_clock")
    Block.config("interval", "Delta between ticks.", default=0.1)
    Block.config("noise", "Jitter.", default=0.02)
    Block.output("clock", "Clock signal.")
    Block.config("length", "Total interval", default=None)

    def init(self):
        self.state.clock = 0

    def update(self):
        self.set_output("clock", "orig: %1.3f" % self.state.clock, timestamp=self.state.clock)

        dt = self.config.interval + np.random.rand() * self.config.noise
        self.state.clock += dt

    def next_data_status(self):
        if self.config.length is not None and self.state.clock > self.config.length:
            return (False, None)
        else:
            return (True, self.state.clock + self.config.interval)


class JitteryDisplay(Block):
    Block.alias("jittery_display")

    Block.input("clock")
    Block.output("rgb")

    def init(self):
        self.plot_generic = PlotGeneric(width=320, height=240, transparent=False, tight=False, keep=True)
        self.first_timestamp = None
        self.plot_anim = PlotAnim()
        self.nframes = 0

    def update(self):
        if self.first_timestamp is None:
            self.first_timestamp = self.get_input_timestamp(0)
        self.time_since_start = self.get_input_timestamp(0) - self.first_timestamp
        self.nframes += 1
        self.output.rgb = self.plot_generic.get_rgb(self.plot)

    def plot(self, pylab):
        self.plot_anim.set_pylab(pylab)
        self.plot_anim.text("clock", 0, 1, "%5.2f" % self.time_since_start)
        self.plot_anim.text("frames", 0, 0.5, "%d" % self.nframes)

        self.plot_anim.text("value", 0, 0.24, self.input.clock)
        pylab.axis((-0.2, 1.1, -0.1, 1.1))



""" 
    This is a set of blocks to read and write logs in HDF_ format.
    
    You need the pytables_ package to be installed.
    
    .. _pytables: http://pytables.org
     
    .. _HDF: http://en.wikipedia.org/wiki/Hierarchical_Data_Format

"""
from procgraph import import_succesful, import_magic

procgraph_info = {
    # List of python packages
    "requires": ["tables", "hdflog"]
}

tables = import_magic(__name__, "tables")
hdflogs = import_magic(__name__, "hdflog")

if import_succesful(hdflogs):
    from .hdfwrite import *
    from .hdfread import *
    from .hdfread_many import *
else:
    from procgraph import logger
#     logger.warning('Could not import hdflogs')

from procgraph import pg_add_this_package_models

pg_add_this_package_models(__file__, __package__)



import string
import traceback

from contracts import contract, describe_value

from procgraph.utils import indent, levenshtein_best_match


class PGException(Exception):
    pass


class BlockWriterError(PGException):
    """ An error by who wrote the block (e.g.: did not define signals)."""

    pass


class ModelWriterError(PGException):
    """ An error by who wrote the model, can be either Syntax or Semantic """

    pass


class SemanticError(ModelWriterError):
    """ A semantic error by who wrote the model spec.
       (and, as a platypus case, when wrong config is passed."""

    def __init__(self, error, element=None):
        # http://web.archiveorange.com/archive/v/jbUwzgEaecaPQftfkITO
        Exception.__init__(self, error, element)

        self.error = error
        if element is not None:
            assert hasattr(element, "where")
        self.element = element

    def __str__(self):
        s = self.error
        if self.element is not None:
            s += "\n" + format_where(self.element).strip()
        return s


class PGSyntaxError(ModelWriterError):
    """ A syntactic error by who wrote the model spec."""

    def __init__(self, error, where=None):
        # http://web.archiveorange.com/archive/v/jbUwzgEaecaPQftfkITO
        Exception.__init__(self, error, where)

        self.error = error
        self.where = where

    def __str__(self):
        s = self.error
        s += "\n\n" + add_prefix(self.where.__str__(), " ")
        return s


class ModelInstantionError(SemanticError):
    pass


class ModelExecutionError(PGException):
    """ 
        Runtime errors, including misuse by the user.
    
    """

    # XXX: code is repeated above
    def __init__(self, error, element=None):
        # http://web.archiveorange.com/archive/v/jbUwzgEaecaPQftfkITO
        Exception.__init__(self, error, element)

        self.error = error
        if element is not None:
            assert hasattr(element, "where")
        self.element = element

    def __str__(self):
        s = string.strip(self.error)
        if self.element is not None:
            s += format_where(self.element)
        return string.strip(s)


class BadMethodCall(ModelExecutionError):
    def __init__(self, method, block, user_exception: str):
        self.method = method
        self.blocks = [block]
        self.user_exception = user_exception

    def __str__(self):
        block = self.blocks[-1]
        s = "User-thrown exception while calling %s() in block %r." % (self.method, block.name)
        for b in self.blocks[::-1]:
            s += "\n- %s %s" % (b, id(b))
        s += "\n" + indent(str(self.user_exception).strip(), "> ")
        return s


class BadInput(ModelExecutionError):
    """ Exception thrown to communicate a problem with one
        of the inputs to the block. """

    def __init__(self, error, block, input_signal):
        # http://web.archiveorange.com/archive/v/jbUwzgEaecaPQftfkITO
        Exception.__init__(self, error, block, input_signal)

        self.block = block
        self.error = error
        self.input_signal = input_signal
        if block is not None:
            self.bad_value = block.get_input(input_signal)
        else:
            self.bad_value = "??? (block not given)"

    def __str__(self):
        if self.block is not None:
            name = self.block.name
        else:
            name = "(unknown)"

        # TODO: add short bad_value
        s = "Bad input %r for block %r: %s" % (self.input_signal, name, self.error)

        if self.block is not None:
            s += "\n value: %s" % describe_value(self.bad_value)

        s += format_where(self.block)
        return s


class BadConfig(ModelExecutionError):
    """ Exception thrown to communicate a problem with one
        of the configuration values passed to the block. """

    def __init__(self, error, block, config):
        Exception.__init__(self, error, block, config)

        self.config = config
        self.error = error
        self.block = block
        self.bad_value = block.config[config]

    def __str__(self):
        if self.block is not None:
            name = self.block.name
        else:
            name = "(unknown)"

        s = "Bad config %r = %r for block %r: %s" % (self.config, self.bad_value, name, self.error)
        s += format_where(self.block)
        return s


# A couple of functions for pretty errors
def aslist(x):
    if isinstance(x, dict):
        x = list(x.keys())
    if x:
        return ", ".join(sorted(x))
    else:
        return "<empty>"


@contract(what="str", x="str", returns="str")
def x_not_found(what, x, iterable):
    """ Shortcut for creating pretty error messages. """
    options = list(iterable)
    if not options:
        msg = "Could not find %s %r; no options available." % (what, x)
        return msg
    guess, _ = levenshtein_best_match(x, options)

    msg = "Could not find %s %r in %d options. (Did you mean %r?)" % (what, x, len(options), guess)
    return msg


def add_prefix(s, prefix):
    result = ""
    for l in s.split("\n"):
        result += prefix + l + "\n"
    return result


def format_where(element_or_block):
    e = element_or_block
    if e is not None:
        if e.where is not None:
            return "\n\n" + add_prefix(e.where.__str__(), " ")
        else:
            return " (no position given)"
    else:
        return " (no element/block given)"



from . import np
from procgraph import Block, ETERNITY
import warnings


__all__ = ["Solid", "solid"]


class Solid(Block):
    Block.alias("solid")
    Block.output("rgb")
    Block.config("width")
    Block.config("height")
    Block.config("color", default=[1, 1, 1])

    def init(self):
        self.info("init")

    def update(self):
        rgb = solid(self.config.width, self.config.height, self.config.color)
        self.info("updating solid ts : %s" % ETERNITY)
        # self.set_output(0, rgb, timestamp=ETERNITY)
        warnings.warn("Eternity is not respected")
        self.set_output(0, rgb, timestamp=0)


def solid(width, height, color):
    rgb = np.zeros((height, width, 3), dtype="uint8")
    for i in range(3):
        rgb[:, :, i] = color[i] * 255
    return rgb



from ..core.model_loader import model_from_string

from .utils import PGTestCase


class DelayedTest(PGTestCase):
    def test_delayed(self):

        model_desc = """
           |generic_in0_out3| -> |identity| -> |identity| -> |block1:identity|
        """
        model = model_from_string(model_desc)

        block1 = model.name2block["block1"]
        self.assertTrue(block1.are_output_signals_defined())

        print(block1)


# XXX: what's the deal here?
#
#    def test_check_definitions(self):
#
#        model_desc = """
#            |DoesNotDefineOutput|
#        """
#
#        self.check_semantic_error(model_desc)
#
#
#        model_desc = """
#            |DoesNotDefineInput|
#        """
#        self.check_semantic_error(model_desc)



import unittest
from numpy.testing import assert_array_almost_equal

from .gradient1d import gradient1d


class TestGradient(unittest.TestCase):
    def testInvalid(self):
        self.assertRaises(ValueError, gradient1d, [0])
        self.assertRaises(ValueError, gradient1d, [0, 0])

    def testSimply(self):
        examples = [([0, 0, 0], [0, 0, 0]), ([1, 1, 1], [0, 0, 0]), ([1, 2, 3], [1, 1, 1])]

        for f, gf in examples:
            g = gradient1d(f)

            assert_array_almost_equal(g, gf)



from .utils import PGTestCase

from ..core.parsing import parse_value
from ..core.parsing_elements import VariableReference

examples = {
    "0": 0,
    "0.0": 0.0,
    "1.0": 1.0,
    "1": 1,
    "[]": [],
    "[1]": [1],
    "[1.0]": [1.0],
    "[1,2]": [1, 2],
    "[1,[2]]": [1, [2]],
    "[[]]": [[]],
    "[{}]": [{}],
    "{}": {},
    "[{a:0}]": [{"a": 0}],
    # XXX
    #'{a:[]}': {"a":[]},
    "{a:0}": {"a": 0},
    "{a:0,b:1}": {"a": 0, "b": 1},
    "{a:b}": {"a": "b"},
    "{a:{b:c}}": {"a": {"b": "c"}},
    """{ indices: [0,170] }""": {"indices": [0, 170]},
    """{ indices: [0,170], theta: [-1,+1], color: 'r', \
            origin: [0,0,0],  max_readings: 5}""": {
        "indices": [0, 170],
        "theta": [-1, +1],
        "color": "r",
        "origin": [0, 0, 0],
        "max_readings": 5,
    },
    """[{ indices: [0,170], theta: [-1,+1], color: 'r', \
      origin: [0,0,0],  max_readings: 5}, { indices: [171,341], \
        theta: [+1,+5], color: 'b', origin: [0,0,0], max_readings: 5}]""": [
        {"indices": [0, 170], "theta": [-1, +1], "color": "r", "origin": [0, 0, 0], "max_readings": 5},
        {"indices": [171, 341], "theta": [+1, +5], "color": "b", "origin": [0, 0, 0], "max_readings": 5},
    ],
    "...": Ellipsis,
    "[1,2,...]": [1, 2, Ellipsis],
    "$v": VariableReference("v"),
    # simple string is a string
    "goodname": "goodname",
}


class SyntaxTest2(PGTestCase):
    def test_numbers(self):
        self.assertEqual(type(parse_value("1")), int)

    def test_numbers2(self):
        self.assertEqual(type(parse_value("1.0")), float)
        # assert isinstance(parse_value('1.0'), float)

    def test_numbers3(self):
        l = parse_value("[1.0]")
        assert isinstance(l[0], float)

    def test_numbers4(self):
        l = parse_value("[1]")
        assert isinstance(l[0], int)

    def testExamples(self):
        for example, expected in list(examples.items()):
            print(("Trying '%s'" % example))
            found = parse_value(example)

            ok = (found == expected) or (str(found) == str(expected))

            if not ok:
                msg = "Parsing: %r\n" % example
                msg += " expected: %r (%s) \n" % (expected, type(expected))
                msg += "    found: %r (%s) \n" % (found, type(found))
                raise Exception(msg)



from contracts import contract
from system_cmd.which import find_executable

__all__ = ["check_programs_existence"]


@contract(returns="dict(str:str)")
def check_programs_existence(programs):
    res = {}
    for p in programs:
        e = find_executable(p)
        if e is None:
            msg = "Cannot find program %r." % p
            raise ValueError(msg)
        else:
            res[p] = e

    return res



from numpy import array, linspace, sin, cos
from procgraph import Block, BadInput
import math


__all__ = ["LaserDotDisplay"]


class LaserDotDisplay(Block):
    """ Produces a plot of a range-finder scan variation (derivative).

    It is a variation of :ref:`block:laser_display`; look there for
    the documentation.

    """

    Block.alias("laser_dot_display")

    Block.config("width", "Width of the resulting image.", default=320)
    Block.config("height", "Height of the resulting image.", default=320)

    Block.config("groups", "How to group and draw the readings. " " (see :ref:`block:laser_display`) ")
    Block.config(
        "title", "By default it displays the signal name." " Set the empty string to disable.", default=None
    )
    Block.config("transparent", "Gives transparent RGBA rather than RGB.", default=False)
    Block.config("R0", "Radius of the readings circle.", default=1)
    Block.config("amp", "Amplitude of the readings crown.", default=0.5)

    Block.input("readings_dot", "Array of float representing array readings.")

    Block.output("image", "A fancy visualization of the laser derivative")

    def update(self):
        y = array(self.input.readings_dot)
        from procgraph_mpl import pylab2rgb, pylab

        if max(abs(y)) > 1:
            raise BadInput(
                "I expect an input normalized in the [-1,1] range;" "min,max: %s,%s " % (min(y), max(y)),
                self,
                "readings_dot",
            )

        f = pylab.figure(frameon=False, figsize=(self.config.width / 100.0, self.config.height / 100.0))

        R0 = self.config.R0
        amp = self.config.amp

        theta = linspace(0, 2 * math.pi, 300)
        pylab.plot(R0 * cos(theta), R0 * sin(theta), "k--")

        for group in self.config.groups:
            indices = group["indices"]
            indices = list(range(indices[0], indices[-1] + 1))
            theta_spec = group["theta"]
            origin = group.get("origin", [0, 0, 0])
            color = group.get("color", "b.")

            N = len(indices)

            theta = linspace(theta_spec[0], theta_spec[-1], N)

            group_y = y[indices]

            r = R0 + amp * group_y

            px = cos(theta + origin[2]) * r
            py = sin(theta + origin[2]) * r

            pylab.plot(-py, px, color)

        M = R0 + amp * 1.2
        pylab.axis([-M, M, -M, M])

        # turn off ticks labels, they don't have meaning
        pylab.setp(f.axes[0].get_xticklabels(), visible=False)
        pylab.setp(f.axes[0].get_yticklabels(), visible=False)

        if self.config.title is not None:
            if self.config.title != "":
                pylab.title(self.config.title, fontsize=10)
        else:
            # We don't have a title ---
            t = self.get_input_signals_names()[0]
            pylab.title(t, fontsize=10)

        self.output.image = pylab2rgb(transparent=self.config.transparent)

        pylab.close(f.number)



from .utils import PGTestCase

good_examples = [
    "u = -1",
    "|GENERATOR| --> result",
    "result --> |SINK|",
    "u -> |FUNCTION| --> result",
    "u ,v -> |FUNCTION| --> result, d",
    "u -> |opa:FUNCTION| ->  result",
    "u --> |op1:FUNCTION op=tan| -> result",
    "u -> |FUNCTION| -> |FUNCTION2| -> result",
    "u -> |FUNCTION| -> a -> |FUNCTION2| -> result",
    "u -> |FUNCTION| -> a -> |FUNCTION2| -> |FUNCTION2| -> result",
    "u -> |FUNCTION| -> a -> |FUNCTION2| ---> a,d --> |FUNCTION2| -> result",
    """  |origin| -> a  -> |block| """,
    """ |constant value=12| -> [0] a  -> |block| """,
    "u = 1",
    """u = 1
u = 2
u =3 """,
    """u = 2
x -> |func2| -> res3""",
    """x -> |func2| 
u = 2""",
    """

x -> |func2| 

u = 2

""",
    " a -> |block| ",
    " a[o] -> |block| ",
    " a.s[o] -> |block| ",
    " [i]a -> |block| ",
    " [i]a[d] -> |block| ",
    " [i]a[d], b -> |block| ",
    " [i]a[d], b[c] -> |block| ",
    " b1.f [input_name] -> |test| -> [t] y [U], z [t] -> |test| -> res",
    """ 
b1.f [input_name] -> |test| -> [t] y [U], z [t] -> |test| -> res
""",
    """
# this is a comment
u = 1
""",
    # I don't think we use it anymore
    # in case, readd:
    # # parameter_list = delimitedList(key_value_pair) ^ OneOrMore(key_value_pair)
    # " |generic in=0,out=2|  ",
    " |generic in=0 out=2|  ",
    # assignment with object
    """ 
g1.in = 2
""",
    # quoted strings
    """
g1.in = "ciao"
""",
    """
g1.in = 'ciao'
""",
    # Triple quotes
    """
g1.in = '''ciao
come va'''
""",
    '''
g1.in = """ciao
come va"""
''',
    """
g1.in = "ci\\"a/o"
""",
    # underscores
    " |output name=x_old|  ",
    # new syntax: allow whitespace
    """
|a| -> |b| \ 
-> |c|
""",
    """
|a| -> |b| -> 
    -> |c|
""",
    """
|a| -> |b|  
    -> |c|
""",
    " (in) -> |a| ",
    " (in) -> in ",
    " a -> |a| -> (out) ",
    " |a| -> (out) ",
    " a -> (out) ",
    " (in) -> |a| -> (out) ",
]


bad_examples = [
    # empty string gives error
    "   \n \n ",
    "u = -1a",
    "|GENERATOR| --> result resu",
    "|GENERATOR GENERATOR| --> result resu",
    "result x --> |SINK|",
    "u -> |FUNCTION| |FUNCTION| --> result",
    # invalid names
    ".u = 2",
    ".u.d = 2",
    "1u = 2",
    # bad quotation
    """ a = "ci"ao" """,
]


class SyntaxTest(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from .utils import PGTestCase

# Note that you have to escape the slash in python strings.
good_examples = [
    """
y --> |block| --> z --> |block| 
""",
    """ # without whitespace after 
y \\
  --> |block| --> z --> |block| 
""",
    """ # with whitespace after 
y \\ 
  --> |block| --> z --> |block| 
""",
    """ 
y --> |block| --> \\
    z --> |block| 
""",
    """ 
y --> |block| --> z \\ 
    --> |block| 
"""
    """ 
y --> |block| --> z --> \\ 
    |block| 
""",
]

bad_examples = []


class TestLineContinuations(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from typing import List, Optional

from contracts import contract, new_contract
from .block_config import resolve_config
from .block_meta import BlockMeta, BlockMetaSugar
from .block_sugar import ConfigProxy, InputProxy, OutputProxy, StateProxy
from .exceptions import BlockWriterError, ModelExecutionError, ModelWriterError

__all__ = [
    "Block",
    "NOT_READY",
    "Generator",
]

NOT_READY = None

new_contract("num_or_id", "int|str")


class Block(BlockMetaSugar, metaclass=BlockMeta):
    def __init__(self, name, config, library):  # @UnusedVariable
        assert isinstance(name, str), "The block name must be a string, not %s" % name.__class__

        self.name = name

        list_of_config = self.__class__.config
        self.__config = resolve_config(list_of_config, config, None)

        # this is an array containing the names/ids
        # example: ["y", 1, 2]
        self.__input_signal_names = None
        # example: {"y":0, "1":1, "2":2}
        self.__input_signal_name2id = None
        # this is an array of Values
        self.__input_signals = None

        # same as above
        self.__output_signal_names = None
        self.__output_signal_name2id = None
        self.__output_signals = None

        # state variables
        self.__states = {}

        # instantiation point
        self.where = None

        # proxies for accessing input, output, and state
        self.input = InputProxy(self)
        self.output = OutputProxy(self)
        self.state = StateProxy(self)
        self.config = ConfigProxy(self)

        # used for debug; hierarchical level
        self.level = 0

        # used for info_once() and error_once(): messages we already sent
        self._msgs_written = set()
        # Used by input_update_available()
        self._input_update_available_last = {}

    def init(self):
        """ Initializes the block.  """
        pass

    def update(self):
        """ Performs the block function. """
        pass

    def finish(self):
        pass

    def cleanup(self):
        pass

    # TODO: move away
    UPDATE_NOT_FINISHED = "update-not-finished"

    # Used during initialization
    def num_input_signals(self):
        assert self.are_input_signals_defined(), "No input defined for %s" % self
        return len(self.__input_signals)

    def num_output_signals(self):
        assert self.are_output_signals_defined(), "No output defined for %s" % self
        return len(self.__output_signals)

    def get_input_signals_names(self):
        """ Returns the list of names of currently defined input signals. """
        assert self.are_input_signals_defined(), "No input defined for %s" % self
        return list(self.__input_signal_names)

    def get_output_signals_names(self):
        """ Returns the list of names of currently defined output signals. """
        assert self.are_output_signals_defined(), "No output defined for %s" % self
        return list(self.__output_signal_names)

    def are_input_signals_defined(self):
        return self.__input_signals is not None

    def are_output_signals_defined(self):
        return self.__output_signals is not None

    def define_input_signals_new(self, signals):
        if not isinstance(signals, list):
            msg = "I expect the parameter to define_input_signals() to be a " "list, instead got a %s: %s" % (
                signals.__class__.__name__,
                signals,
            )
            raise BlockWriterError(msg)

        # Make sure unique
        if len(set(signals)) != len(signals):
            msg = "Not unique set of signals names: %s" % signals
            raise BlockWriterError(msg)

        # reset structures
        self.__input_signal_names = []
        self.__input_signals = []
        self.__input_signal_name2id = {}
        for i, s in enumerate(signals):
            if not isinstance(s, str):
                msg = "Invalid list of names for input: %s " % signals
                raise BlockWriterError(msg)

            self.__input_signal_names.append(str(s))
            self.__input_signal_name2id[str(s)] = i
            self.__input_signals.append(Value(None, NOT_READY))

    @contract(returns="bool")
    def all_input_signals_ready(self):
        """ Returns True if all input signals are ready. """
        for value in self.__input_signals:
            if value.timestamp == NOT_READY:
                return False
        return True

    #     @contract(signal='num_or_id', returns='bool')
    def input_signal_ready(self, signal):
        """
            Return True if the signal is ready, meaning that we have
            at least one value.
        """
        res = bool(self.__get_input_struct(signal).timestamp != NOT_READY)
        # print('ready? %s %s' % (signal, res))
        return res

    @contract(signal="num_or_id", returns="bool")
    def input_update_available(self, signal):
        """
            Return True if the signal was updated since the last
            time this function was called with the same argument.

            Returns False if it's not ready.
        """

        if not self.input_signal_ready(signal):
            return False
        current = self.get_input_timestamp(signal)
        track = self._input_update_available_last

        if not signal in track:
            track[signal] = current
            return True

        last_seen = track[signal]

        update_available = last_seen != current

        track[signal] = current

        return bool(update_available)

    def define_output_signals_new(self, signals):
        if not isinstance(signals, list):
            msg = "I expect the parameter to define_output_signals()" + " to be a list, got a %s: %s" % (
                signals.__class__.__name__,
                signals,
            )
            raise BlockWriterError(msg)

        # reset structures
        self.__output_signal_names = []
        self.__output_signals = []
        self.__output_signal_name2id = {}
        for i, s in enumerate(signals):
            if not isinstance(s, str):
                msg = "Invalid list of names for output: %s; " " I expect strings." % signals
                raise BlockWriterError(msg)

            self.__output_signal_names.append(str(s))
            self.__output_signal_name2id[str(s)] = i
            self.__output_signals.append(Value(None, NOT_READY))

    def get_config(self, conf):
        if not conf in self.__config:
            msg = "For block %s: could not find parameter %r in config %s." % (self, conf, self.__config)
            raise ModelExecutionError(msg, self)

        return self.__config[conf]

    def set_state(self, varname, value):
        """ Can be called during init() and update(). """
        # TODO: add check and exception
        self.__states[varname] = value

    def get_state(self, varname):
        """ Can be called during init() and update() ."""
        if not varname in self.__states:
            raise ValueError("No such state variable %r" % varname)
        return self.__states[varname]

    def get_state_vars(self):
        """ Returns a list of the names for the state variables. """
        # TODO: remove this?
        return list(self.__state.keys())

    # Functions that can be called during runtime
    def set_output(self, num_or_id, value, timestamp=None):
        """ Sets an output value. If timestamp is omitted, it
            will default to the maximum of the input signals timestamp. """
        if timestamp is None:
            if len(self.__input_signals) == 0:
                msg = "%s: setting %s: timestamp not specified and no inputs" % (self, num_or_id)
                raise Exception(msg)

            timestamp = max(self.get_input_signals_timestamps())

        output_struct = self.__get_output_struct(num_or_id)
        output_struct.value = value
        output_struct.timestamp = timestamp

    def from_outside_set_input(self, num_or_id, value, timestamp):
        """ Sets an input value. This is used from outside, not
        from the block. (This is overloaded by Model) """
        if timestamp is None:
            msg = "Setting input %r to %s with None timestamp" % (num_or_id, timestamp)
            raise ValueError(msg)
        input_struct = self.__get_input_struct(num_or_id)
        input_struct.value = value
        input_struct.timestamp = timestamp

    @contract(returns="tuple(float, *)")
    def get_input_ts_and_value(self, num_or_id):
        """ Gets the timestamp and value of an input signal. """
        if not self.input_signal_ready(num_or_id):
            msg = "Cannot get_input_ts_and_value(%r): input not ready." % num_or_id
            raise ValueError(msg)

        input_struct = self.__get_input_struct(num_or_id)
        value = input_struct.value
        ts = input_struct.timestamp
        return ts, value

    @contract(num_or_id="num_or_id")
    def get_input(self, num_or_id):
        """ Gets the value of an input signal. """
        if not self.input_signal_ready(num_or_id):
            msg = "Cannot get_input(%r): input not ready." % num_or_id
            raise ValueError(msg)

        input_struct = self.__get_input_struct(num_or_id)
        return input_struct.value

    @contract(num_or_id="num_or_id")
    def get_input_timestamp(self, num_or_id):
        """
            Gets the timestamp of an input signal (None if it
            has never been received).
        """
        if not self.input_signal_ready(num_or_id):
            msg = "Cannot get_input_timestamp(%r): input not ready." % num_or_id
            raise ValueError(msg)
        input_struct = self.__get_input_struct(num_or_id)
        return input_struct.timestamp

    @contract(num_or_id="num_or_id")
    def get_output_timestamp(self, num_or_id):
        """ Gets the timestamp of an output signal. """
        output_struct = self.__get_output_struct(num_or_id)
        return output_struct.timestamp

    @contract(num_or_id="num_or_id")
    def get_output(self, num_or_id):
        """ Gets the value of an output signal. """
        output_struct = self.__get_output_struct(num_or_id)
        return output_struct.value

    @contract(num_or_id="num_or_id")
    def __get_input_struct(self, num_or_id):
        """ Returns a reference to the Value structure of the given
            input signal.
        """
        if not self.is_valid_input_name(num_or_id):
            msg = "Unknown input name %r." % str(num_or_id)
            raise ModelWriterError(msg, self)

        if isinstance(num_or_id, str):
            # convert from name to number
            num_or_id = self.__input_signal_name2id[num_or_id]
        return self.__input_signals[num_or_id]

    @contract(num_or_id="num_or_id")
    def __get_output_struct(self, num_or_id):
        """ Returns a reference to the Value structure of the given
            output signal.
        """
        if not self.is_valid_output_name(num_or_id):
            msg = "Unknown output name %r." % str(num_or_id)
            raise ModelWriterError(msg, self)

        if isinstance(num_or_id, str):
            # convert from name to number
            num_or_id = self.__output_signal_name2id[num_or_id]
        return self.__output_signals[num_or_id]

    @contract(num_or_id="num_or_id")
    def is_valid_input_name(self, num_or_id):
        """ Checks that num_or_id (string or int) is a valid handle
            for one of the signals. """
        assert self.are_input_signals_defined(), "No input defined for %s (checking %r)" % (self, num_or_id)

        if isinstance(num_or_id, str):
            return num_or_id in self.__input_signal_name2id
        if isinstance(num_or_id, type(0)):
            return num_or_id < len(self.__input_signals)
        raise ValueError("Invalid input name %r" % num_or_id)

    @contract(num_or_id="num_or_id")
    def canonicalize_input(self, num_or_id):
        """ Converts the signal spec (either string or id) to string
            (useful because more user-friendly). """
        assert self.is_valid_input_name(num_or_id), "%s: %r is not a valid input name." % (self, num_or_id)
        if isinstance(num_or_id, str):
            return num_or_id
        if isinstance(num_or_id, type(0)):
            return self.__input_signal_names[num_or_id]
        raise ValueError("Invalid input name %r" % num_or_id)

    @contract(num_or_id="num_or_id")
    def is_valid_output_name(self, num_or_id):
        """ Checks that num_or_id (string or int) is a valid handle
            for one of the signals. """
        assert self.are_output_signals_defined(), "No output defined for %s (checking %r)" % (self, num_or_id)

        if isinstance(num_or_id, str):
            return num_or_id in self.__output_signal_name2id
        if isinstance(num_or_id, type(0)):
            return num_or_id < len(self.__output_signals)

        raise ValueError("Invalid output type %r" % num_or_id)

    @contract(num_or_id="num_or_id")
    def canonicalize_output(self, num_or_id):
        """ Converts the signal spec (either string or id) to string
            (useful because more user-friendly). """
        assert self.is_valid_output_name(num_or_id), "%s: %r is not a valid output name." % (self, num_or_id)
        if isinstance(num_or_id, str):
            return num_or_id
        if isinstance(num_or_id, type(0)):
            return self.__output_signal_names[num_or_id]
        raise ValueError(f"Invalid output name {num_or_id!r}")

    def get_output_signals_timestamps(self) -> List[Optional[float]]:
        """ Returns a list of the output values timestamps. """
        return [x.timestamp for x in self.__output_signals]

    def get_input_signals_timestamps(self) -> List[Optional[float]]:
        """ Returns a list of the input signals timestamps. """
        return [x.timestamp for x in self.__input_signals]

    def get_input_signals_values(self) -> List[object]:
        """ Returns a list of the input signals values. """
        return [x.value for x in self.__input_signals]

    def __repr__(self):
        s = "B:%s:%s(" % (self.__class__.__name__, self.name)
        s += self.get_io_repr()
        s += ")"
        return s

    def get_io_repr(self):
        """ Returns a representation of io ports for this block. """
        s = ""
        if self.are_input_signals_defined():
            if self.__input_signals:
                s += "in:%s" % ",".join(self.__input_signal_names)
            else:
                s += "in:/"
        else:
            s += "in:?"

        s += ";"
        if self.are_output_signals_defined():
            if self.__output_signals:
                s += "out:%s" % ",".join(self.__output_signal_names)
            else:
                s += "out:/"
        else:
            s += "out:?"
        return s

    def _log_prefix(self):
        return "%s%s:" % (">" * self.level, self.name)

    def info_once(self, msg):
        if not msg in self._msgs_written:
            self.info(msg)
            self._msgs_written.add(msg)

    def error_once(self, msg):
        if not msg in self._msgs_written:
            self.error(msg)
            self._msgs_written.add(msg)

    def info(self, s):
        """ Writes an info message. """
        from .visualization import info as pg_info

        pg_info("%s%s" % (self._log_prefix(), s))

    def warning(self, s):
        """ Writes an info message. """
        from .visualization import warning as pg_warning

        pg_warning("%s%s" % (self._log_prefix(), s))

    def debug(self, s):
        """ Writes a debug message. """
        from .visualization import debug as pg_debug

        pg_debug("%s%s" % (self._log_prefix(), s))

    def error(self, s):
        """ Writes a debug message. """
        from .visualization import error as pg_error

        pg_error("%s%s" % (self._log_prefix(), s))


class Generator(Block):
    # TODO: change interface

    def next_data_status(self):
        """
        This is complicated but necessary. Do you have another value?

        - Yes, and it will be of this timestamp. (I can see it from the log)

        - No, this generator has finished.

        - Yes, but this is realtime and it does not depend on me.
          For example, I'm waiting for the next sensor data.
          Ask me later.

        In those cases, we return:

            (True, timestamp)

            (False, None)

            (True, None)

        """
        raise NotImplementedError('"next_data_status" was not implement.')


class Value(object):
    """
        timestamp = ETERNITY     constant
        timestamp = NOT_READY = (None)         signal not ready
    """

    def __init__(self, value, timestamp):
        self.value = value
        self.timestamp = timestamp



from procgraph import Block


class LowPass(Block):
    """ Implements simple low-pass filtering. 
    
    Formula used: ::
    
        y[k] = alpha * u[k] + (1-alpha) * y[k-1]
    
    """

    # TODO: make a serious low-pass block
    Block.alias("low_pass")

    Block.config("alpha", "Innovation rate")

    Block.input("value", "Any numpy array.")
    Block.output("lowpass", "The lowpass version.")

    def init(self):
        self.state.y = None

    def update(self):
        u = self.input[0]
        alpha = self.config.alpha

        if self.state.y is None:
            self.state.y = u
        else:
            self.state.y = self.state.y * (1 - alpha) + alpha * u

        self.output[0] = self.state.y



from procgraph import Block


class ToFile(Block):
    """ Prints the input line by line to a given file."""

    Block.alias("to_file")

    Block.config("file", "File to write.")

    Block.input("values", "Anything you wish to print to file.")

    def init(self):
        self.file = open(self.config.file, "w")

    def update(self):
        s = str(self.input[0])
        self.file.write(s)
        self.file.write("\n")
        self.file.flush()



""" 
    An example package for ProcGraph that shows how to organize your code. 

    This is the documentation string for the package. Like all docstrings,
    it consists of a short summary (above) and a longer description (this.)
    
"""


# Specify all the meta-information in the 'procgraph_info' structure.
# You can omit all of this at first, then come back and fill it in once you
# want your package to properly install on other people's machines.

procgraph_info = {
    # purely informational
    "author": "Andrea Censi <andrea@cds.caltech.edu>",
    "project_url": "http://procgraph.org",
    "scm": "git@",
    # List of python packages
    "requires": [
        # try to add this
        # 'unexistent',
        ("pickle", ("cPickle", "pickle")),
    ],
    # List of external software needed
    "requires_executables": ["git"],
    # The following are more advanced, and used only for generating
    # rich cross-referenced documentation.
    "sphinx_base_url": "http://andreacensi.github.com/procgraph/",
    "scm_base_url": "https://github.com/AndreaCensi/procgraph/blob/master/",
}


from procgraph import import_magic

# If pickle is installed, it will be a reference to it, otherwise a
# shadow object which will throw when you actually try to use it.
pickle = import_magic(__name__, "pickle")

# procgraph will let you import unexistent modules
# unexistent = import_magic(__name__, 'unexistent')


# now, import the modules that defines the blocks
from . import example0_simplest_block



from procgraph import Block
from procgraph.core.registrar_other import register_model_spec
from procgraph.scripts.pgmain import pg
import numpy as np


__all__ = ["pg_quick_animation"]


class GenericPlot(Block):

    Block.alias("GenericPlot")

    Block.config("plotfunc", "Must be called with plotfunc(pylab, frame)")
    Block.config("width", "Image dimension", default=320)
    Block.config("height", "Image dimension", default=240)
    Block.config("transparent", "If true, outputs a RGBA image instead of RGB.", default=False)
    Block.config("tight", 'Uses "tight" option for creating png (Matplotlib>=1.1).', default=False)

    Block.input("tick")
    Block.output("rgb")

    def init(self):
        from procgraph_mpl import PlotGeneric

        self.plot_generic = PlotGeneric(
            width=self.config.width,
            height=self.config.height,
            transparent=self.config.transparent,
            tight=self.config.tight,
        )
        self.state.frame = 0

    def update(self):
        frame = self.state.frame

        plotfunc = self.config.plotfunc

        def f(pylab):
            return plotfunc(pylab, frame)

        self.output.rgb = self.plot_generic.get_rgb(f)

        self.state.frame += 1


register_model_spec(
    """
--- model GenericWrap
''' This is a simple test for :ref:`block:hdfread_many`. '''
config width
config height
config out
config plotfunc
config nframes 
config fps

import procgraph_mpl.examples

|clock interval=1.0 length=$nframes| -> clock
clock -> |GenericPlot plotfunc=$plotfunc width=$width height=$height| -> rgb

rgb -> |enc:mencoder fps=$fps file=$out |
"""
)


def pg_quick_animation(plotfunc, nframes, width, height, out, fps):
    config = dict(width=width, height=height, out=out, nframes=nframes, plotfunc=plotfunc, fps=fps)

    config["enc.firstpass_bitrate"] = 10 * 1000 * 1000

    pg("GenericWrap", config=config)


if __name__ == "__main__":

    def bounce(pylab, frame):
        t = frame * 0.2
        t0 = t
        t1 = t + 2
        x = np.linspace(t0, t1, 1000)
        y = np.cos(x)
        pylab.plot(x, y)
        pylab.axis((t0, t1, -1.2, +1.2))

    out = "bounce.mp4"

    pg_quick_animation(bounce, nframes=10, width=320, height=240, out=out, fps=10)



import numpy as np
from procgraph import COMPULSORY
from procgraph import simple_block

from .pil_conversions import Image_from_array


__all__ = ["resize"]


@simple_block
def pil_zoom(value, factor=COMPULSORY):
    """ Zooms by a given factor """
    # TODO: RGBA?
    shape = value.shape[:2]
    shape2 = (int(factor * shape[0]), int(factor * shape[1]))
    height, width = shape2
    image = Image_from_array(value)
    image = image.resize((width, height))
    result = np.asarray(image.convert("RGB"))
    return result


@simple_block
def resize(value, width=None, height=None):
    """ 
        Resizes an image.
        
        You should pass at least one of ``width`` or ``height``.
        
        :param value: The image to resize.
        :type value: image
        
        :param width: Target image width.
        :type width: int,>0
        
        :param height: Target image height.
        :type height: int,>0

        :return: image: The image as a numpy array.
        :rtype: rgb
    """
    H, W = value.shape[:2]

    if width is None and height is None:
        msg = "You should pass at least one of width or height."
        raise ValueError(msg)

    if width is None and height is not None:
        width = int((height * 1.0 / H) * W)
    elif height is None and width is not None:
        height = int((width * 1.0 / W) * H)

    if width == W and height == H:
        # print('want: %s have: %s -- No resize necessary' % (value.shape, (width, height)))
        return value.copy()

    image = Image_from_array(value)
    # TODO: RGBA?
    image = image.resize((width, height))
    result = np.asarray(image.convert("RGB"))

    assert result.shape[0] == height
    assert result.shape[1] == width

    # ratio1 = value.shape[0] * 1.0 / value.shape[1]
    # ratio2 = result.shape[0] * 1.0 / result.shape[1]
    # print('ratio1 : %.2f  result %.2f' % (ratio1, ratio2))

    return result



from .metadata import ffmpeg_get_metadata, read_extra_metadata_for
from .timestamps import TIMESTAMP_FIELD, timestamp_from_iso
from contracts import contract, raise_wrapped
from procgraph.utils import CmdException, system_cmd_result
import math
import os


@contract(returns="dict(str:*)")
def pg_video_info(filename, intolerant=True):
    """
          Returns a dictionary, with fields:
        
            metadata
            
            width, height
            
            fps* (might fail)
            length* (might fail)
            
            timestamp # start timestamp
                        
            extra_mencoder_info
                
        Tries to read precise timestamp from metadata; otherwise 
        it tries from the .timestamps file; otherwise from the file's mtime.
        
        Mplayer might fail to identify FPS and length. If intolerant=True,
        an error is raised, otherwise they are set to None. 
    """
    if not os.path.exists(filename):
        msg = "File %r does not exist." % filename
        raise ValueError(msg)

    info = mplayer_identify(filename, intolerant=intolerant)

    info["metadata"] = ffmpeg_get_metadata(filename)

    extra_md = read_extra_metadata_for(filename)

    info["metadata"].update(extra_md)

    precise = info["metadata"].get(TIMESTAMP_FIELD, None)
    if precise is None:
        # logger.info('No precise timestamp in metadata found for %s' % filename)
        timestamp = os.path.getmtime(filename)
    else:
        # logger.info('Precise timestamp found for %s' % filename)
        timestamp = timestamp_from_iso(precise)

    timestamps = filename + ".timestamps"
    if os.path.exists(timestamps):
        # logger.info('Reading timestamps from %r.' % timestamps)
        f = open(timestamps)
        line = f.readline()
        # print('frist line: %s' % line)
        timestamp = float(line)
        # print('timestamp: %s' % timestamp)

    info["timestamp"] = timestamp

    return info


def mplayer_identify(filename, intolerant=True):
    """
        Returns a dictionary, with fields:
        
            width, height
            fps *
            length * 
            
            extra_mencoder_info
            
        Mplayer might fail to identify FPS and length. If intolerant=True,
        an error is raised, otherwise they are set to None.
    
    """

    # need at least 1 frame otherwise sometimes the video aspect is not known
    args = "mplayer -identify -vo null -ao null -frames 1".split() + [filename]

    try:

        try:
            res = system_cmd_result(
                ".",
                args,
                display_stdout=False,
                display_stderr=False,
                raise_on_error=True,
                capture_keyboard_interrupt=False,
            )
        except CmdException:
            raise

    except Exception as e:
        raise_wrapped(Exception, e, "Could not identify movie", filename=filename)

    try:
        output = res.stdout
        return parse_mplayer_info_output(output, intolerant=intolerant)
    except Exception as e:
        raise_wrapped(
            Exception, e, "Could not identify movie", cmd=" ".join(args), filename=filename, output=output
        )


def parse_mplayer_info_output(output, intolerant=True):
    keys = ["ID_VIDEO_WIDTH", "ID_VIDEO_HEIGHT", "ID_VIDEO_FPS", "ID_LENGTH", "ID_VIDEO_ASPECT"]
    id_width, id_height, id_fps, id_length, id_video_aspect = keys

    info = {}
    for line in output.split("\n"):
        if line.startswith("ID_"):
            key, value = line.split("=", 1)
            try:  # interpret numbers if possible
                value = eval(value)
            except:
                pass
            if key in info:
                # print('overriding %r=%r from %r' % (key,value,info[key]))
                pass
            info[key] = value

    for k in keys:
        if not k in info:
            msg = "Could not find key %r in properties %s." % (k, sorted(info.keys()))
            raise Exception(msg)

    if id_length == 0:
        msg = "I could not find find the length of this video."
        raise Exception(msg)

    res = {}
    res["width"] = info[id_width]
    res["height"] = info[id_height]
    if isinstance(info[id_fps], str):
        msg = "Invalid fps %r:\n" % info[id_fps]
        if intolerant:
            raise ValueError(msg)
        else:
            print(msg)  # XXX
        res["fps"] = None
    else:
        res["fps"] = float(info[id_fps])

    res["length"] = info[id_length]

    if not res["length"] > 0:
        if intolerant:
            msg = "invalid length %r" % res["length"]
            raise ValueError(msg)
        else:
            res["length"] = None
    #     check('float,>0', res['length'])

    res["extra_mencoder_info"] = info

    w_over_h = res["width"] * 1.0 / res["height"]

    dar = info[id_video_aspect] * 1.0
    if dar == 0:
        # dar = 1.0
        dar = w_over_h

    res["dar"] = dar
    #     print('dar', dar)

    # dar/sar = width/height

    # print('w_over_h', w_over_h)
    # sar = dar / w_over_h
    sar = res["dar"] / w_over_h
    #     print('sar', sar)
    if math.fabs(1.0 - sar) < 0.01:
        sar = 1.0
    res["sar"] = sar
    return res



from .utils import PGTestCase

from procgraph import Block, BlockWriterError


def same_name_mistake_config():
    class MyBlock(Block):
        Block.config("x", "description")
        Block.config("x", "description 2", default=True)


def same_name_mistake_input():
    class MyBlock(Block):
        Block.input("x")
        Block.input("x")


def same_name_mistake_output():
    class MyBlock(Block):
        Block.output("x")
        Block.output("x")


def bad_mixing_1():
    class MyBlock(Block):
        Block.output("x")
        Block.output_is_variable()


def bad_mixing_2():
    class MyBlock(Block):
        Block.output_is_variable()
        Block.output("x")


def bad_mixing_3():
    class MyBlock(Block):
        Block.input_is_variable()
        Block.input("x")


def bad_mixing_4():
    class MyBlock(Block):
        Block.input("x")
        Block.input_is_variable()


def bad_mixing_5():
    class MyBlock(Block):
        Block.input("x")
        Block.output_is_variable()


def bad_mixing_6():
    class MyBlock(Block):
        Block.output_is_variable()
        Block.input("x")


def bad_mixing_7():
    class MyBlock(Block):
        Block.output_is_variable()


def good_mixing_1():
    class MyBlockA(Block):
        Block.input_is_variable()
        Block.output_is_variable()


def good_mixing_2():
    class MyBlockB(Block):
        Block.input_is_variable()
        Block.output("only one")


def good_mixing_3():
    class MyBlockC(Block):
        Block.input_is_variable()


class SyntaxTestMultiple(PGTestCase):
    def test_same_name_mistake(self):
        """ Test that we detect when a input, output, config name
            is repeated. 
        """
        self.assertRaises(BlockWriterError, same_name_mistake_config)
        self.assertRaises(BlockWriterError, same_name_mistake_input)
        self.assertRaises(BlockWriterError, same_name_mistake_output)

    def test_variable_inputs(self):
        self.assertRaises(BlockWriterError, bad_mixing_1)
        self.assertRaises(BlockWriterError, bad_mixing_2)
        self.assertRaises(BlockWriterError, bad_mixing_3)
        self.assertRaises(BlockWriterError, bad_mixing_4)
        # XXX: it's late, not sure if this is correct or not
        # self.assertRaises(BlockWriterError, bad_mixing_5)
        # self.assertRaises(BlockWriterError, bad_mixing_6)
        # self.assertRaises(BlockWriterError, bad_mixing_7)

        good_mixing_1()
        good_mixing_2()
        good_mixing_3()

    def test_ok(self):
        class MyBlockOK(Block):
            Block.config("x", "description")
            Block.config("y", "description 2", default=True)
            Block.config("z")
            Block.input("x")
            Block.input("y")
            Block.output("x")

        self.assertEqual(len(MyBlockOK.config), 3)
        self.assertEqual(len(MyBlockOK.input), 2)
        self.assertEqual(len(MyBlockOK.output), 1)



from procgraph import Block

__all__ = ["Retime", "RewriteTimestamps"]


class Retime(Block):
    """ 
        Multiplies timestamps by give factor
    """

    Block.alias("retime")

    Block.config("factor", "Factor")

    Block.input("x")
    Block.output("y")

    def init(self):
        pass

    def update(self):
        value = self.get_input(0)
        t = self.get_input_timestamp(0)
        t_ = t * self.config.factor
        self.set_output(0, value, t_)


class RewriteTimestamps(Block):
    """ 
        Retims the timestamps equally spaced.
        
        [0, interval, interval*2, interval*3, ...]
    """

    Block.alias("rewrite_timestamps")

    Block.config("interval", "interval")

    Block.input("x")
    Block.output("y")

    def init(self):
        self.i = 1

    def update(self):
        value = self.get_input(0)
        t = self.i * self.config.interval
        self.i += 1
        self.set_output(0, value, t)



import numpy as np
from procgraph import simple_block
import struct
import numpy


@simple_block
def ros2python(msg):
    """ Converts a ROS message to a Python object. """
    # FIXME: this is a hack
    #    print msg.__class__
    #    if isinstance(msg, String):
    #        return msg.data
    #    else:
    #        print('Unknown type: %s' % msg.__class__)
    #        return msg

    return msg.data


@simple_block
def ros_scan2python(scan):
    #    print scan.__dict__.keys()
    return np.array(scan.ranges)


def rgb_from_imgmsg(image):
    im, _, _ = imgmsg_to_pil(image)
    pix = np.asarray(im).astype(np.uint8)
    return pix


# @PydevCodeAnalysisIgnore
# #Code from:
# http://nullege.com/codes/show/src%40c%40m%40cmu-ros-pkg-HEAD%40trunk%40posedetectiondb%40src%40ParseMessages.py/75/sensor_msgs.msg.Image/python
#

import PIL.Image  # @UnresolvedImport

# import numpy
# import roslib  # @UnresolvedImport @UnusedImport
# import rospy  # @UnresolvedImport @UnusedImport
# import sensor_msgs.msg  # @UnresolvedImport
# from sensor_msgs.msg import CompressedImage  # @UnusedImport @UnresolvedImport
# import struct

#
@simple_block
def ros2rgb(msg):
    # print msg.__class__.__name__ # _sensor_msgs__CompressedImage
    if "CompressedImage" in msg.__class__.__name__:
        # print 'format: %s' % msg.format
        # if msg.format == 'jpeg':
        #     data = msg.data
        #
        #     from PIL import ImageFile  # @UnresolvedImport
        #     parser = ImageFile.Parser()
        #     parser.feed(data)
        #     res = parser.close()
        #     print res
        return rgb_from_pil(pil_from_CompressedImage(msg))
    else:
        return rgb_from_imgmsg(msg)


def pil_from_CompressedImage(msg):
    from PIL import ImageFile  # @UnresolvedImport

    parser = ImageFile.Parser()
    parser.feed(msg.data)
    res = parser.close()
    return res


def rgb_from_pil(im):
    return np.asarray(im).astype(np.uint8)


#
# Number of channels used by a PIL image mode
def imgmsg_to_pil(
    rosimage,
    encoding_to_mode={
        "mono8": "L",
        "8UC1": "L",
        "8UC3": "RGB",
        "rgb8": "RGB",
        "bgr8": "RGB",
        "rgba8": "RGBA",
        "bgra8": "RGBA",
        "bayer_rggb": "L",
        "bayer_gbrg": "L",
        "bayer_grbg": "L",
        "bayer_bggr": "L",
        "yuv422": "YCbCr",
        "yuv411": "YCbCr",
    },
    PILmode_channels={"L": 1, "RGB": 3, "RGBA": 4, "YCbCr": 3},
):
    conversion = "B"
    channels = 1
    if rosimage.encoding.find("32FC") >= 0:
        conversion = "f"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("64FC") >= 0:
        conversion = "d"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("8SC") >= 0:
        conversion = "b"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("8UC") >= 0:
        conversion = "B"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("16UC") >= 0:
        conversion = "H"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("16SC") >= 0:
        conversion = "h"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("32UC") >= 0:
        conversion = "I"
        channels = int(rosimage.encoding[-1])
    elif rosimage.encoding.find("32SC") >= 0:
        conversion = "i"
        channels = int(rosimage.encoding[-1])
    else:
        if rosimage.encoding.find("rgb") >= 0 or rosimage.encoding.find("bgr") >= 0:
            channels = 3

    data = struct.unpack(
        (">" if rosimage.is_bigendian else "<")
        + "%d" % (rosimage.width * rosimage.height * channels)
        + conversion,
        rosimage.data,
    )

    if conversion == "f" or conversion == "d":
        dimsizes = [rosimage.height, rosimage.width, channels]
        imagearr = numpy.array(255 * I, dtype=numpy.uint8)  # @UndefinedVariable
        im = PIL.Image.frombuffer(
            "RGB" if channels == 3 else "L", dimsizes[1::-1], imagearr.tostring(), "raw", "RGB", 0, 1
        )
        if channels == 3:
            im = PIL.Image.merge("RGB", im.split()[-1::-1])
        return im, data, dimsizes
    else:
        mode = encoding_to_mode[rosimage.encoding]
        step_size = PILmode_channels[mode]
        dimsizes = [rosimage.height, rosimage.width, step_size]
        im = PIL.Image.frombuffer(mode, dimsizes[1::-1], rosimage.data, "raw", mode, 0, 1)
        if mode == "RGB":
            im = PIL.Image.merge("RGB", im.split()[-1::-1])
        return im, data, dimsizes


#
# def pil_to_imgmsg(image, encodingmap={'L': 'mono8', 'RGB': 'rgb8', 'RGBA': 'rgba8', 'YCbCr': 'yuv422'},
#                         PILmode_channels={'L': 1, 'RGB': 3, 'RGBA': 4, 'YCbCr': 3}):
#     rosimage = sensor_msgs.msg.Image()
#     # adam print 'Channels image.mode: ',PILmode_channels[image.mode]
#     rosimage.encoding = encodingmap[image.mode]
#     (rosimage.width, rosimage.height) = image.size
#     rosimage.step = PILmode_channels[image.mode] * rosimage.width
#     rosimage.data = image.tostring()
#     return rosimage
#
# def numpy_to_imgmsg(image, stamp=None):
#     rosimage = sensor_msgs.msg.Image()
#     rosimage.height = image.shape[0]
#     rosimage.width = image.shape[1]
#     if image.dtype == numpy.uint8:
#         rosimage.encoding = '8UC%d' % image.shape[2]
#         rosimage.step = image.shape[2] * rosimage.width
#         rosimage.data = image.ravel().tolist()
#     else:
#         rosimage.encoding = '32FC%d' % image.shape[2]
#         rosimage.step = image.shape[2] * rosimage.width * 4
#         rosimage.data = numpy.array(image.flat, dtype=numpy.float32).tostring()
#     if stamp is not None:
#         rosimage.header.stamp = stamp
#     return rosimage



from procgraph import Block
from datetime import datetime

__all__ = ["Info"]


class Info(Block):
    """ 
        Prints more compact information about the inputs 
        than :ref:`block:print`.
    
        For numpy arrays it prints their shape and dtype 
        instead of their values. 
        
    """

    Block.alias("info")
    Block.input_is_variable("Signals to describe.", min=1)

    def init(self):
        self.first = {}
        self.counter = {}

    def update(self):
        # Just copy the input to the output
        for i in range(self.num_input_signals()):
            name = self.canonicalize_input(i)
            val = self.get_input(i)
            ts = self.get_input_timestamp(i)

            if ts is None:
                continue

            if not i in self.first:
                self.first[i] = ts
                self.counter[i] = 0
            friendly = ts - self.first[i]
            #             if isinstance(val, numpy.ndarray):
            #                 s = "%s %s" % (str(val.shape), str(val.dtype))
            #             else:
            s = str(val)
            if len(s) > 40:
                s = s[:40]
            s = s.replace("\n", "|")
            date = datetime.fromtimestamp(ts).isoformat(" ")[:-4]
            ts = "%.2f" % ts
            self.debug("%s (%8.2fs) %12s %5d  %s" % (date, friendly, name, self.counter[i], s))
            self.counter[i] += 1



from .conversions.containers import do_quickstart
from .conversions.metadata import get_ffmpeg_metadata_args
from .conversions.vcodecs import get_x264_encoder_params, get_prores_encoder_params
from system_cmd import system_cmd_result, CmdException
import os


def validate_args(filename, out, my_ext):
    """ Returns out """

    basename, _ = os.path.splitext(filename)

    if out is None:
        out = basename + my_ext

    if os.path.splitext(out)[1] != my_ext:
        msg = "I expect a %r as out (%r)" % (my_ext, out)
        raise ValueError(msg)

    if os.path.splitext(filename)[1] == my_ext:
        msg = "Need a file that does not end in %r (%r)" % (my_ext, filename)
        raise ValueError(msg)

    if not os.path.exists(filename):
        msg = "Input filename does not exist (%r)" % filename
        raise ValueError(msg)

    if os.path.exists(out):
        os.unlink(out)

    return out


def convert_to_mov_prores(filename, out=None, quiet=True, profile=3, qv=None, timestamp=None, metadata={}):
    my_ext = ".mov"
    out = validate_args(filename, out, my_ext)

    cmds = ["ffmpeg"]
    cmds += ["-y"]
    cmds += ["-i", filename]
    cmds += ["-f", "mov"]
    cmds += ["-an"]  # no audio
    cmds += get_prores_encoder_params(profile, qv)
    cmds += get_ffmpeg_metadata_args(metadata, timestamp)
    cmds += [out]

    try:
        system_cmd_result(
            ".",
            cmds,
            display_stdout=not quiet,
            display_stderr=not quiet,
            raise_on_error=True,
            capture_keyboard_interrupt=False,
        )
    except CmdException:
        if os.path.exists(out):
            os.unlink(out)
        raise

    assert os.path.exists(out)


def convert_to_mkv_mp4(filename, out=None, quiet=True, crf=18, preset="medium", timestamp=None, metadata={}):
    """
        Converts to a matrioska file with all metadata.
        
    """
    my_ext = ".mkv"
    out = validate_args(filename, out, my_ext)

    cmds = ["ffmpeg"]
    cmds += ["-y"]
    cmds += ["-i", filename]
    cmds += get_x264_encoder_params(crf, preset)
    cmds += get_ffmpeg_metadata_args(metadata, timestamp)
    cmds += [out]

    try:
        system_cmd_result(
            ".",
            cmds,
            display_stdout=not quiet,
            display_stderr=not quiet,
            raise_on_error=True,
            capture_keyboard_interrupt=False,
        )
    except CmdException:
        if os.path.exists(out):
            os.unlink(out)
        raise

    assert os.path.exists(out)


def convert_to_mp4(filename, mp4=None, quiet=True, crf=18, preset="medium", timestamp=None, metadata={}):
    """ 
        Creates a web-ready mp4 using ffmpeg.
    
        Needs either qtquickstart (from the python package) or qt-quickstart from ffmpeg.
        
        On ubuntu 12.04: need 'sudo apt-get install x264 libavcodec-extra-53'
        to install necessary codecs.
        You can see a list of supported presets by using: 'x264 --help'.
    
        
        (other packages that might be helpful: # libavdevice-extra-52  libavfilter-extra-0 
         libavformat-extra-52 libavutil-extra-49 libpostproc-extra-51 libswscale-extra-0)
       
    """
    my_ext = ".mp4"

    if os.path.splitext(mp4)[1] != my_ext:
        msg = "I expect a %r as out (%r)" % (my_ext, mp4)
        raise ValueError(msg)

    if os.path.splitext(filename)[1] == my_ext:
        msg = "Need a file that does not end in %r (%r)" % (my_ext, filename)
        raise ValueError(msg)

    basename, _ = os.path.splitext(filename)

    if mp4 is None:
        mp4 = basename + my_ext

    # need .mp4 at the end otherwise ffmpeg gets confused
    tmp = basename + ".mp4.firstpass.mp4"

    if not os.path.exists(filename):
        raise ValueError("File does not exist: %s" % filename)

    #    if (os.path.exists(mp4) and
    #        (os.path.getmtime(mp4) > os.path.getmtime(filename))):
    #        return

    if os.path.exists(tmp):
        os.unlink(tmp)
    if os.path.exists(mp4):
        os.unlink(mp4)

    cmds = ["ffmpeg"]
    cmds += ["-y"]
    cmds += ["-i", filename]
    cmds += get_x264_encoder_params(crf, preset)
    cmds += get_ffmpeg_metadata_args(metadata, timestamp)
    cmds += [tmp]

    # print cmds

    try:
        system_cmd_result(
            ".",
            cmds,
            display_stdout=not quiet,
            display_stderr=not quiet,
            raise_on_error=True,
            capture_keyboard_interrupt=False,
        )
    except CmdException:
        if os.path.exists(tmp):
            os.unlink(tmp)
        raise

    do_quickstart(source=tmp, target=mp4)

    if os.path.exists(tmp):
        os.unlink(tmp)



from ...core.registrar_other import simple_block
from ...core.block import Block


class Minus(Block):
    """
        Implements the difference of two signals, taking care of overflows.

        Because that is rarely the semantics you want to give them.
    """

    Block.alias("-")
    Block.input("x", "First signal")
    Block.input("y", "Second signal")
    Block.output("x_minus_y", "Result of x - y")
    Block.config(
        "safe",
        "Whether to use safe promotions. " "If not specified, we will do it but warn once.",
        default=None,
    )

    Block.config("cases", "Promotion rules", default={"uint8": "int16", "uint16": "int32", "uint32": "int64"})

    def init(self):
        self.warned = False

    def update(self):
        # TODO:  check_array_same_shape(self, 0, 1)
        inputs = [self.get_input(0), self.get_input(1)]

        warned_now = False
        for i in [0, 1]:
            if inputs[i] is None:
                return
            dtype_string = str(inputs[i].dtype)
            if dtype_string in self.config.cases:
                nexti = self.config.cases[dtype_string]

                if not self.warned:
                    self.info(
                        "Warning: promoting signal %r from %r to %r."
                        % (self.canonicalize_input(i), dtype_string, nexti)
                    )
                    warned_now = True
                inputs[i] = inputs[i].astype(nexti)

        self.warned = self.warned or warned_now

        result = inputs[0] - inputs[1]

        self.set_output(0, result)


@simple_block(alias="+")
def plus(x, y):
    """
        Sum of two signals.

        :param x: First signal.
        :param y: Second signal.
        :return: sum: Sum of the two signals.
    """
    return x + y


@simple_block(alias="*")
def product(x, y):
    """
        Product of two signals.

        :param x: First signal.
        :param y: Second signal.
        :return: product: Product of the two signals.

    """
    return x + y


@simple_block(alias="/")
def ratio(x, y) -> float:
    """
        Ratio of two signals.

        :param x: First signal.
        :param y: Second signal.
        :return: ratio: First signal divided by the second.
    """
    return x / y



import sys
from os import environ as env

try:
    from termcolor import colored as termcolor_colored  # @UnresolvedImport
except:
    sys.stderr.write('procgraph can make use of the package "termcolor". ' "Please install it.\n")

    def termcolor_colored(x, color=None, on_color=None, attrs=None):  # @UnusedVariable
        """ emulation of the termcolor interface """
        return x


def colored(x, color=None, on_color=None, attrs=None):
    colorize = True
    # TODO: no colorize during tests
    if colorize:  # @UndefinedVariable
        return termcolor_colored(x, color, on_color, attrs)
    else:
        return x


try:
    from setproctitle import setproctitle  # @UnresolvedImport @UnusedImport
except:
    sys.stderr.write('procgraph can make use of the package "setproctitle". ' "Please install it.\n")

    def setproctitle(x):
        """ emulation of the setproctitle interface """
        pass


screen_columns = None


def get_screen_columns():
    module = sys.modules[__name__]
    if module.screen_columns is None:
        max_x, max_y = getTerminalSize()  # @UnusedVariable
        module.screen_columns = max_x

    return module.screen_columns


def getTerminalSize():
    """
    max_x, max_y = getTerminalSize()
    """
    import os

    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            import struct

            cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
        except:
            return None
        return cr

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (env["LINES"], env["COLUMNS"])
        except:
            cr = (25, 80)
    return int(cr[1]), int(cr[0])


# These should be kwargs to
#    colored(s, color=None, on_color=None, attrs=None
color_warning = dict(color="yellow")
color_error = dict(color="red")
color_user_error = dict(color="red")
color_info = dict(color="green")
color_debug = dict(color="magenta")


# TODO: use logging
if False:

    def warning(string):
        write_message(string, lambda x: "pg: " + colored(x, **color_warning))

    def error(string):
        write_message(string, lambda x: "pg: " + colored(x, **color_error))

    def user_error(string):
        write_message(string, lambda x: "pg: " + colored(x, **color_user_error))

    def info(string):
        write_message(string, lambda x: "pg: " + colored(x, **color_info))

    def debug(string):
        write_message(string, lambda x: "pg: " + colored(x, **color_debug))

    def write_message(string, formatting):

        sys.stdout.flush()
        string = str(string)

        # clean_console_line(sys.stderr)
        lines = string.split("\n")
        if len(lines) == 1:
            sys.stderr.write(formatting(lines[0]) + "\n")
        else:
            for i, l in enumerate(lines):  # @UnusedVariable
                # if i == 1:
                #    l = '- ' + l
                # else:
                #    l = '  ' + l
                sys.stderr.write(formatting(l) + "\n")

        sys.stderr.flush()


else:

    prefix = ""
    from procgraph import logger

    def warning(string):
        fmt = lambda x: prefix + colored(x, **color_warning)
        logger.warning(fmt(string))

    def error(string):
        fmt = lambda x: prefix + colored(x, **color_error)
        logger.error(fmt(string))

    def user_error(string):
        fmt = lambda x: prefix + colored(x, **color_user_error)
        logger.error(fmt(string))

    def info(string):
        fmt = lambda x: prefix + colored(x, **color_info)
        logger.info(fmt(string))

    def debug(string):
        fmt = lambda x: prefix + colored(x, **color_debug)
        logger.info(fmt(string))


def semantic_warning(error, element):
    msg = str(error) + "\n" + str(element.where)
    warning(msg)



from procgraph import Block
from procgraph_images.copied_from_reprep import skim_top_and_bottom, skim_top

from numpy import maximum, minimum
import numpy


class OrganicScale(Block):
    """ A (almost failed!) attempt to scale a signal into [-1,1] 
        according to the history. 
        
        This one is a mess.
    """

    Block.alias("organic_scale")

    Block.input("value")
    Block.output("value_scaled")

    Block.config("skim", default=5)
    Block.config("skim_hist", default=5)
    Block.config("hist", "How many steps of history to use.", default=100)
    Block.config("tau", default=0.1)

    def init(self):
        self.state.max = []
        self.state.M = None

    def update(self):
        y = numpy.array(self.input.value)
        # first skim it
        y = skim_top_and_bottom(y, self.config.skim)
        # put the max in the history
        self.state.max.insert(0, max(abs(y)))
        # prune the history
        self.state.max = self.state.max[: self.config.hist]
        # get the history percent
        M = max(skim_top(numpy.array(self.state.max), self.config.skim_hist))

        if self.state.M is None:
            self.state.M = M
        else:
            self.state.M += self.config.tau * (M - self.state.M)

        y = minimum(y, self.state.M)
        y = maximum(y, -self.state.M)

        y = y / self.state.M
        self.output.value_scaled = y



import numpy as np
from procgraph import Block
from contracts import contract


class DepthBuffer(Block):

    Block.alias("depth_buffer")

    Block.input("rgba")
    Block.output("rgba")
    Block.output("line")
    Block.output("depth")

    def init(self):
        self.depth = None

    def update(self):
        rgba = self.input.rgba
        if self.depth is None:
            H, W = rgba.shape[0:2]
            self.depth = np.zeros((H, W))
            self.depth.fill(0)

        d = get_depth(rgba)

        mask = rgba[:, :, 3] > 0
        closer = np.logical_and(self.depth < d, mask)
        farther = np.logical_not(closer)

        self.depth[closer] = d

        rgba = rgba.copy()
        rgba[farther, 3] = 0

        with_line = rgba[:, :, 0:3].copy()
        with_line[d, :, 0] = 255
        with_line[d, :, 1] = 55

        depth = self.depth.copy()
        depth[depth == 0] = np.nan
        self.output.rgba = rgba
        self.output.line = with_line
        self.output.depth = depth


@contract(rgba="array[HxWx4](uint8)", returns="float,>=0,<=H-1")
def get_depth(rgba) -> float:
    alpha = rgba[:, :, 3]
    H, _ = alpha.shape

    a = 0
    w = 0
    for i in range(H):
        line = alpha[i, :].astype("float32")
        a += i * np.sum(line)
        w += np.sum(line)
    a = a / w
    return a



__all__ = [
    "PlotAnim",
]


class PlotAnim(object):
    def __init__(self):
        self.handle_line = {}
        self.handle_text = {}
        self.pylab = None

    def set_pylab(self, pylab):
        self.pylab = pylab

    def assert_pylab_given(self):
        if self.pylab is None:
            msg = "Please call set_pylab() before plotting."
            raise ValueError(msg)

    def plot(self, name, x, y, *args, **kwargs):
        self.assert_pylab_given()
        if not name in self.handle_line:
            (handle,) = self.pylab.plot(x, y, *args, **kwargs)
            self.handle_line[name] = handle
        else:
            handle = self.handle_line[name]
            handle.set_data(x, y)

    def text(self, name, x, y, text, *args, **kwargs):
        self.assert_pylab_given()
        if not name in self.handle_text:
            handle = self.pylab.text(x, y, text, *args, **kwargs)
            self.handle_text[name] = handle
        else:
            handle = self.handle_text[name]
            handle.set_text(text)



# from collections import namedtuple
import re

# Arg = namedtuple('Arg', 'name type')


class Arg(object):
    def __init__(self, desc=None, type=None):  # @ReservedAssignment
        if desc is not None:
            desc = desc.strip()
        self.desc = desc
        if type is not None:
            type = type.strip()  # @ReservedAssignment
        self.type = type

    def __eq__(self, other):
        return self.type == other.type and self.desc == other.desc

    def __repr__(self):
        return "Arg(%r,%r)" % (self.desc, self.type)


class DocStringInfo(object):
    def __init__(self, docstring=None, params={}, returns=[]):
        self.docstring = docstring
        self.params = params
        self.returns = returns
        # self.vars
        # self.raises

    def __eq__(self, other):
        return (
            self.docstring == other.docstring
            and self.params == other.params
            and self.returns == other.returns
        )

    def __str__(self):
        s = "DocString:\n"
        s += " params: %s\n" % self.params
        s += " returns: %s" % self.returns
        return s


def parse_annotations(docstring, keys, empty=False):
    """ Returns docstring_without, dictionary.
        If empty specified, will look for empty statements, and give integers
        for names. 
    """
    assert docstring is not None

    found = {}

    for key in keys:
        if empty:
            regexp = "^\s*:\s*%s\s*:\s*(?P<desc>.*?)\s*$" % key
        else:
            regexp = "^\s*:\s*%s\s+(?P<name>\w*?)\s*:\s*(?P<desc>.*?)\s*$" % key
        regexp = re.compile(regexp, re.MULTILINE)

        def replace(match):
            d = match.groupdict()

            if empty:
                name = len(found)
            else:
                name = d["name"]

            found[name] = d["desc"]

            return ""

        docstring = regexp.sub(repl=replace, string=docstring)

    return docstring, found


def parse_docstring_annotations(docstring):
    assert docstring is not None

    param_keys = ["param", "parameter", "arg", "argument", "key", "keyword"]
    type_keys = ["type"]
    return_keys = ["returns", "return"]
    rtype_keys = ["rtype"]
    # TODO: document state?
    # var_keys = ['var', 'ivar', 'cvar']
    # raises, raise, except, exception

    docstring, params_ann = parse_annotations(docstring, param_keys, False)
    docstring, types_ann = parse_annotations(docstring, type_keys, False)
    docstring, returns_ann = parse_annotations(docstring, return_keys, True)
    docstring, rtype_ann = parse_annotations(docstring, rtype_keys, True)

    params = {}
    names = set(list(params_ann.keys()) + list(types_ann.keys()))
    for name in names:
        params[name] = Arg(params_ann.get(name, None), types_ann.get(name, None))

    returns = []
    for i in range(max(len(returns_ann), len(rtype_ann))):
        returns.append(Arg(returns_ann.get(i, None), rtype_ann.get(i, None)))

    return DocStringInfo(docstring, params=params, returns=returns)
